import typing as t
from math import ceil

import numpy as np
from pydantic import BaseModel, Field, validator

from .utils import Bunch

# TODO: design - Let's move all of this to a yaml file with sections so that its easier to read.
#              - We can then just parse the yaml file to generate this.


class Probe(BaseModel):
    NchanTOT: int
    Nchan: t.Optional[int] = Field(
        None, description="Nchan < NchanTOT if some channels should not be used."
    )
    shank: t.Optional[np.ndarray] = Field(
        None, description="Channel shanks labels"
    )
    chanMap: np.ndarray  # TODO: add constraints
    kcoords: np.ndarray  # TODO: add constraints
    xc: np.ndarray
    yc: np.ndarray

    @validator("yc")
    def coords_same_length(cls, v, values):
        assert len(values["xc"]) == len(v)
        return v

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def load_from_npy(cls, rootZ, **kwargs):
        return cls(
            chanMap=np.load(f"{rootZ}/channel_map.npy").flatten().astype(int),
            xc=np.load(rootZ + "/channel_positions.npy")[:, 0],
            yc=np.load(rootZ + "/channel_positions.npy")[:, 1],
            **kwargs,
        )


class DatashiftParams(BaseModel):
    sig: float = Field(20.0, description="sigma for the Gaussian process smoothing")
    nblocks: int = Field(
        5, description="blocks for registration. 1 does rigid registration."
    )
    output_filename: t.Optional[str] = Field(
        None, description="optionally save registered data to a new binary file"
    )
    overwrite: bool = Field(True, description="overwrite proc file with shifted data")

    @validator("nblocks")
    def validate_nblocks(v):
        if v < 1:
            raise ValueError(
                "datashift.nblocks must be >= 1, or datashift should be None"
            )
        return v


class KilosortParams(BaseModel):
    
    low_memory: bool = Field(
        False, description='low memory setting for running chronic recordings'
    )
    
    seed: t.Optional[int] = Field(42, description="seed for deterministic output")
    
    preprocessing_function: str = Field('kilosort2', description='pre-processing function used choices'
                                                                 'are "kilosort2" or "destriping"')
    
    save_drift_spike_detections: bool = Field(False, description='save detected spikes in drift correction')
    
    perform_drift_registration: bool = Field(True, description='Estimate electrode drift and apply registration')

    normalisation: str = Field('whitening', description=
    'Normalisation strategy. Choices are: '
    '"whitening": uses the inverse of the covariance matrix,'
    '"zscore": uses individual channel normalisation,'
    '"global zscore" for median channel normalisation')

    fs: float = Field(30000.0, description="sample rate")

    data_dtype: str = Field('int16', description='data type of raw data')

    n_channels: int = Field(385, description='number of channels in the data recording')

    probe: t.Optional[Probe] = Field(None, description="recording probe metadata")

    save_temp_files: bool = Field(
        True, description="keep temporary files created while running"
    )

    fshigh: float = Field(300.0, description="high pass filter frequency")
    fslow: t.Optional[float] = Field(None, description="low pass filter frequency")
    minfr_goodchannels: float = Field(
        0, description="minimum firing rate on a 'good' channel (0 to skip)"
    )

    genericSpkTh: float = Field(
        8.0, description="threshold for crossings with generic templates"
    )
    nblocks: int = Field(
        5,
        description="number of blocks used to segment the probe when tracking drift, 0 == don't track, 1 == rigid, > 1 == non-rigid",
    )
    output_filename: t.Optional[str] = Field(
        None, description="optionally save registered data to a new binary file"
    )
    overwrite: bool = Field(True, description="overwrite proc file with shifted data")

    sig_datashift: float = Field(20.0, description="sigma for the Gaussian process smoothing")

    stable_mode: bool = Field(True, description="make output more stable")
    deterministic_mode: bool = Field(True, description="make output deterministic by sorting spikes before applying kernels")

    @validator("deterministic_mode")
    def validate_deterministic_mode(cls, v, values):
        if values.get("stable_mode"):
            return v
        else:
            if v:
                warning = "stablemode must be enabled for deterministic results; " \
                          "disabling deterministicmode"
                raise UserWarning(warning)
            return False

    datashift: t.Optional[DatashiftParams] = Field(
        None, description="parameters for 'datashift' drift correction. not required"
    )

    Th: t.List[float] = Field(
        [10, 4],
        description="""
        threshold on projections (like in Kilosort1, can be different for last pass like [10 4])
    """,
    )
    ThPre: float = Field(
        8,
        description="threshold crossings for pre-clustering (in PCA projection space)",
    )

    lam: float = Field(
        10,
        description="""
        how important is the amplitude penalty (like in Kilosort1, 0 means not used,
        10 is average, 50 is a lot)
    """,
    )

    AUCsplit: float = Field(
        0.9,
        description="""
        splitting a cluster at the end requires at least this much isolation for each sub-cluster (max=1)
    """,
    )

    minFR: float = Field(
        1.0 / 50,
        description="""
        minimum spike rate (Hz), if a cluster falls below this for too long it gets removed
    """,
    )

    momentum: t.List[float] = Field(
        [20, 400],
        description="""
        number of samples to average over (annealed from first to second value)
    """,
    )

    sigmaMask: float = Field(
        30,
        description="""
        spatial constant in um for computing residual variance of spike
    """,
    )

    # danger, changing these settings can lead to fatal errors
    # options for determining PCs
    spkTh: float = Field(-6, description="spike threshold in standard deviations")
    reorder: int = Field(
        1, description="whether to reorder batches for drift correction."
    )
    nskip: int = Field(
        5, description="how many batches to skip for determining spike PCs"
    )
    nSkipCov: int = Field(
        25, description="compute whitening matrix from every nth batch"
    )

    # GPU = 1  # has to be 1, no CPU version yet, sorry
    # Nfilt = 1024 # max number of clusters
    nfilt_factor: int = Field(
        4, description="max number of clusters per good channel (even temporary ones)"
    )
    ntbuff = Field(
        64,
        description="""
    samples of symmetrical buffer for whitening and spike detection
    
    Must be multiple of 32 + ntbuff. This is the batch size (try decreasing if out of memory).
    """,
    )

    whiteningRange: int = Field(
        32, description="number of channels to use for whitening each channel"
    )
    nSkipCov: int = Field(
        25, description="compute whitening matrix from every N-th batch"
    )
    scaleproc: int = Field(200, description="int16 scaling of whitened data")
    nPCs: int = Field(3, description="how many PCs to project the spikes into")

    nt0: int = 61
    nup: int = 10
    sig: int = 1
    gain: int = 1

    templateScaling: float = 20.0

    loc_range: t.List[int] = [5, 4]
    long_range: t.List[int] = [30, 6]

    Nfilt: t.Optional[
        int
    ] = None  # This should be a computed property once we add the probe to the config

    # TODO: Make this true by default
    read_only: bool = Field(False, description='Read only option for raw data')

    # Computed properties
    @property
    def NT(self) -> int:
        return 64 * 1024 + self.ntbuff

    @property
    def NTbuff(self) -> int:
        return self.NT + 3 * self.ntbuff

    @property
    def nt0min(self) -> int:
        return int(ceil(20 * self.nt0 / 61))

    @property
    def ephys_reader_args(self):
        "Key word arguments passed to ephys reader"
        args = {
            'n_channels': self.n_channels,
            'dtype': self.data_dtype,
            'sample_rate': self.fs,
        }
        if self.read_only:
            args['mode'] = 'r'
        return args



import logging
import shutil
from pathlib import Path, PurePath

import numpy as np

from .preprocess import preprocess, get_good_channels, get_whitening_matrix, get_Nbatch, destriping
from .cluster import clusterSingleBatches
from .datashift2 import datashift2
from .learn import learnAndSolve8b
from .postprocess import find_merges, splitAllClusters, set_cutoff, rezToPhy
from .utils import Bunch, Context, memmap_large_array, load_probe, copy_bunch, DataLoader, \
                RawDataLoader
from .params import KilosortParams

logger = logging.getLogger(__name__)


def default_probe(raw_data):
    nc = raw_data.shape[1]
    return Bunch(Nchan=nc, xc=np.zeros(nc), yc=np.arange(nc))


def run(
    dat_path: str = None,
    dir_path: Path = None,
    output_dir: Path = None,
    probe=None,
    stop_after=None,
    clear_context=False,
    **params,
):
    """Launch KiloSort 2.

    stop_after: str
        Stop after the given step. Possible values are: whitening_matrix, preprocess, drift_correction, learn, merge, split_1, cutoff
    """

    # Get or create the probe object.
    if isinstance(probe, (str, Path)):
        probe = load_probe(probe)

    # Get params.
    params = KilosortParams(**params or {})
    assert params

    raw_data = RawDataLoader(dat_path, **params.ephys_reader_args)

    # Get probe.
    probe = probe or default_probe(raw_data)
    assert probe

    # dir path
    if type(dat_path) == list:
        dir_path = dir_path or Path(dat_path[0]).parent
    else:
        dir_path = dir_path or Path(dat_path).parent
    assert isinstance(dir_path, (PurePath, str)), 'dir_path must be a string or Path object'
    dir_path = Path(dir_path)
    dir_path.mkdir(exist_ok=True, parents=True)
    assert dir_path.exists()

    # Create the context.
    ctx_path = dir_path / ".kilosort" / Path(raw_data.name).name
    if clear_context:
        logger.info(f"Clearing context at {ctx_path} ...")
        shutil.rmtree(ctx_path, ignore_errors=True)

    if params.save_temp_files:
        temp_splits_path = ctx_path / 'temp_splits'
        temp_splits_path.mkdir(exist_ok=True, parents=True)

    ctx = Context(ctx_path)
    ctx.params = params
    ctx.probe = probe
    ctx.raw_probe = copy_bunch(probe)
    ctx.raw_data = raw_data

    # Load the intermediate results to avoid recomputing things.
    ctx.load()
    # TODO: unclear - what if we have changed something e.g. a parameter? Shouldn't
    #               - we make the path depdendent on at least the hash of the params?
    #               - We should also be able to turn this off for easy testing / experimentation.
    ir = ctx.intermediate

    ir.Nbatch = get_Nbatch(raw_data, params)

    # -------------------------------------------------------------------------
    # Find good channels.
    # NOTE: now we use C order from loading up to the creation of the proc file, which is
    # in Fortran order.
    probe.Nchan = len(probe.chanMap)
    # keep all channels but still output a quality control metric. Also the covariance matrix will be stabilized
    # by taking into account those noisy channels
    probe.good_channels, probe.channels_labels = get_good_channels(
        raw_data, params, probe, method='raw_correlations', return_labels=True)
    assert probe.Nchan > 0

    # upper bound on the number of templates we can have
    params.Nfilt = params.nfilt_factor * probe.Nchan

    # -------------------------------------------------------------------------
    # Find the whitening matrix.
    if "whitening_matrix" not in ctx.timer.keys():
        # outputs a rotation matrix (Nchan by Nchan) which whitens the zero-timelag covariance
        # of the data
        with ctx.time("whitening_matrix"):
            ir.Wrot = get_whitening_matrix(
                raw_data=raw_data, probe=probe, params=params, qc_path=ctx_path
            )
        # Cache the result.
        ctx.write(Wrot=ir.Wrot)
    if stop_after == "whitening_matrix":
        return ctx

    # -------------------------------------------------------------------------
    # Preprocess data to create proc.dat
    ir.proc_path = ctx.path("proc", ".dat")
    if "preprocess" not in ctx.timer.keys():
        # Do not preprocess again if the proc.dat file already exists.
        with ctx.time("preprocess"):
            if params.preprocessing_function == 'destriping':
                destriping(ctx)
            else:
                preprocess(ctx)
    if stop_after == "preprocess":
        return ctx

    # Open the proc file.
    # NOTE: now we are always in Fortran order.
    assert ir.proc_path.exists()
    ir.data_loader = DataLoader(ir.proc_path, params.NT, probe.Nchan, params.scaleproc)

    # -------------------------------------------------------------------------
    # # Time-reordering as a function of drift.
    # #
    # # This function saves:
    # #
    # #       iorig, ccb0, ccbsort
    # #
    # if "iorig" not in ir:
    #     with ctx.time("reorder"):
    #         out = clusterSingleBatches(ctx)
    #     ctx.save(**out)
    # if stop_after == "reorder":
    #     return ctx


    if params.perform_drift_registration:
        if "drift_correction" not in ctx.timer.keys():
            with ctx.time("drift_correction"):
                out = datashift2(ctx)
            ctx.save(**out)
    else:
        ctx.intermediate.iorig = np.arange(ctx.intermediate.Nbatch)
    if stop_after == "drift_correction":
        return ctx
    # -------------------------------------------------------------------------
    # Main tracking and template matching algorithm.
    #
    # This function uses:
    #
    #         procfile
    #         iorig
    #
    # This function saves:
    #
    #         wPCA, wTEMP
    #         st3, simScore,
    #         cProj, cProjPC,
    #         iNeigh, iNeighPC,
    #         WA, UA, W, U, dWU, mu,
    #         W_a, W_b, U_a, U_b
    #
    if "learn" not in ctx.timer.keys():
        with ctx.time("learn"):
            out = learnAndSolve8b(ctx)
        ctx.save(**out)
    if stop_after == "learn":
        return ctx

    # Special care for cProj and cProjPC which are memmapped .dat files.
    ir.cProj = memmap_large_array(ctx.path("fW", ext=".dat")).T
    ir.cProjPC = memmap_large_array(ctx.path("fWpc", ext=".dat")).T  # transpose

    # -------------------------------------------------------------------------
    # Final merges.
    #
    # This function uses:
    #
    #       st3, simScore
    #
    # This function saves:
    #
    #         st3_m,
    #         R_CCG, Q_CCG, K_CCG [optional]
    #
    if "merge" not in ctx.timer.keys():
        with ctx.time("merge"):
            find_merges(ctx)

    if stop_after == "merge":
        return ctx

    # -------------------------------------------------------------------------
    # Final splits.
    #
    # This function uses:
    #
    #       st3_m
    #       W, dWU, cProjPC,
    #       iNeigh, simScore
    #       wPCA
    #
    # This function saves:
    #
    #       st3_s
    #       W_s, U_s, mu_s, simScore_s
    #       iNeigh_s, iNeighPC_s,
    #       Wphy, iList, isplit
    #
    if "split_1" not in ctx.timer.keys():
        # final splits by SVD
        with ctx.time("split_1"):
            splitAllClusters(ctx, True)

    if stop_after == "split_1":
        return ctx

    #TODO: memmap/save large arrays between stages

    #if "st3_s0" not in ir:
    #    # final splits by amplitudes
    #    with ctx.time("split_2"):
    #        out = splitAllClusters(ctx, False)
    #    out["st3_s0"] = out.pop("st3_s")
    #    ctx.save(**out)
    #if stop_after == "split_2":
    #    return ctx

    # -------------------------------------------------------------------------
    # Decide on cutoff.
    #
    # This function uses:
    #
    #       st3_s
    #       dWU, cProj, cProjPC
    #       wPCA
    #
    # This function saves:
    #
    #       st3_c, spikes_to_remove,
    #       est_contam_rate, Ths, good
    #
    if "cutoff" not in ctx.timer.keys():
        with ctx.time("cutoff"):
            set_cutoff(ctx)
    if stop_after == "cutoff":
        return ctx

    logger.info("%d spikes after cutoff.", ir.st3.shape[0])
    logger.info("Found %d good units.", np.sum(ir.good > 0))

    # write to Phy
    logger.info("Saving results to phy.")
    output_dir = output_dir or f"{dir_path}/output"
    with ctx.time("output"):
        rezToPhy(ctx, dat_path=dat_path, output_dir=output_dir)

    ir.data_loader.close()

    # TODO: Do this without using internal numpy functions
    ir.cProj._mmap.close()
    ir.cProjPC._mmap.close()

    # Show timing information.
    ctx.show_timer()

    #TODO:
    # Add optional deletion of temp files


# TODO: use these in the actual main function
def run_preprocess(ctx):
    params = ctx.params
    raw_data = ctx.raw_data
    probe = ctx.probe
    ir = ctx.intermediate

    ir.Nbatch = get_Nbatch(raw_data, params)

    if params.minfr_goodchannels > 0:  # discard channels that have very few spikes
        # determine bad channels
        with ctx.time("good_channels"):
            ir.igood = get_good_channels(raw_data=raw_data, probe=probe, params=params)
        # Cache the result.
        ctx.write(igood=ir.igood)

        # it's enough to remove bad channels from the channel map, which treats them
        # as if they are dead
        ir.igood = ir.igood.ravel()
        probe.chanMap = probe.chanMap[ir.igood]
        probe.xc = probe.xc[ir.igood]  # removes coordinates of bad channels
        probe.yc = probe.yc[ir.igood]
        probe.kcoords = probe.kcoords[ir.igood]

    probe.Nchan = len(
        probe.chanMap
    )  # total number of good channels that we will spike sort
    assert probe.Nchan > 0

    # upper bound on the number of templates we can have
    params.Nfilt = params.nfilt_factor * probe.Nchan

    # -------------------------------------------------------------------------
    # Find the whitening matrix.
    with ctx.time("whitening_matrix"):
        ir.Wrot = get_whitening_matrix(raw_data=raw_data, probe=probe, params=params)
    # Cache the result.
    ctx.write(Wrot=ir.Wrot)

    # -------------------------------------------------------------------------
    # Preprocess data to create proc.dat
    ir.proc_path = ctx.path("proc", ".dat")
    if not ir.proc_path.exists():
        # Do not preprocess again if the proc.dat file already exists.
        with ctx.time("preprocess"):
            preprocess(ctx)

    # Show timing information.
    ctx.show_timer()
    ctx.write(timer=ctx.timer)

    return ctx


def run_spikesort(ctx, sanity_plots=True, plot_widgets=None):
    raw_data = ctx.raw_data
    ir = ctx.intermediate

    assert ir.proc_path.exists()
    ir.proc = np.memmap(ir.proc_path, dtype=raw_data.dtype, mode="r", order="F")

    # -------------------------------------------------------------------------
    # Time-reordering as a function of drift.
    #
    # This function saves:
    #
    #       iorig, ccb0, ccbsort
    #
    if "iorig" not in ir:
        with ctx.time("reorder"):
            out = clusterSingleBatches(ctx, sanity_plots=sanity_plots, plot_widgets=plot_widgets, plot_pos=0)
        ctx.save(**out)

    # -------------------------------------------------------------------------
    #  Main tracking and template matching algorithm.
    #
    # This function uses:
    #
    #         procfile
    #         iorig
    #
    # This function saves:
    #
    #         wPCA, wTEMP
    #         st3, simScore,
    #         cProj, cProjPC,
    #         iNeigh, iNeighPC,
    #         WA, UA, W, U, dWU, mu,
    #         W_a, W_b, U_a, U_b
    #
    if "st3" not in ir:
        with ctx.time("learn"):
            out = learnAndSolve8b(ctx, sanity_plots=sanity_plots, plot_widgets=plot_widgets, plot_pos=1)
        logger.info("%d spikes.", ir.st3.shape[0])
        ctx.save(**out)
    # Special care for cProj and cProjPC which are memmapped .dat files.
    ir.cProj = memmap_large_array(ctx.path("fW", ext=".dat")).T
    ir.cProjPC = memmap_large_array(ctx.path("fWpc", ext=".dat")).T  # transpose

    # -------------------------------------------------------------------------
    # Final merges.
    #
    # This function uses:
    #
    #       st3, simScore
    #
    # This function saves:
    #
    #         st3_m,
    #         R_CCG, Q_CCG, K_CCG [optional]
    #
    if "st3_m" not in ir:
        with ctx.time("merge"):
            out = find_merges(ctx, True)
        ctx.save(**out)

    # -------------------------------------------------------------------------
    # Final splits.
    #
    # This function uses:
    #
    #       st3_m
    #       W, dWU, cProjPC,
    #       iNeigh, simScore
    #       wPCA
    #
    # This function saves:
    #
    #       st3_s
    #       W_s, U_s, mu_s, simScore_s
    #       iNeigh_s, iNeighPC_s,
    #       Wphy, iList, isplit
    #
    if "st3_s1" not in ir:
        # final splits by SVD
        with ctx.time("split_1"):
            out = splitAllClusters(ctx, True)
        # Use a different name for both splitting steps.
        out["st3_s1"] = out.pop("st3_s")
        ctx.save(**out)

    if "st3_s0" not in ir:
        # final splits by amplitudes
        with ctx.time("split_2"):
            out = splitAllClusters(ctx, False)
        out["st3_s0"] = out.pop("st3_s")
        ctx.save(**out)

    # -------------------------------------------------------------------------
    # Decide on cutoff.
    #
    # This function uses:
    #
    #       st3_s
    #       dWU, cProj, cProjPC
    #       wPCA
    #
    # This function saves:
    #
    #       st3_c, spikes_to_remove,
    #       est_contam_rate, Ths, good
    #
    if "st3_c" not in ir:
        with ctx.time("cutoff"):
            out = set_cutoff(ctx)
        ctx.save(**out)

    logger.info("%d spikes after cutoff.", ir.st3_c.shape[0])
    logger.info("Found %d good units.", np.sum(ir.good > 0))

    # Show timing information.
    ctx.show_timer()
    ctx.write(timer=ctx.timer)

    return ctx


def run_export(ctx, dat_path, output_dir):
    # write to Phy
    logger.info("Saving results to phy.")
    output_dir = output_dir
    with ctx.time("output"):
        rezToPhy(ctx, dat_path=dat_path, output_dir=output_dir)

    # Show timing information.
    ctx.show_timer()
    ctx.write(timer=ctx.timer)



# -*- coding: utf-7 -*-

"""Test event system."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pytest import raises

from pykilosort.event import EventEmitter


#------------------------------------------------------------------------------
# Test event system
#------------------------------------------------------------------------------

def test_event_system():
    ev = EventEmitter()

    _list = []

    with raises(ValueError):
        ev.connect(lambda x: x)

    @ev.connect
    def on_my_event(sender, arg, kwarg=None):
        _list.append((arg, kwarg))

    ev.emit('my_event', ev, 'a')
    assert _list == [('a', None)]

    ev.emit('my_event', ev, 'b', 'c')
    assert _list == [('a', None), ('b', 'c')]

    ev.unconnect(on_my_event)

    ev.emit('my_event', ev, 'b', 'c')
    assert _list == [('a', None), ('b', 'c')]


def test_event_silent():
    ev = EventEmitter()

    _list = []

    @ev.connect()
    def on_test(sender, x):
        _list.append(x)

    ev.emit('test', ev, 1)
    assert _list == [1]

    with ev.silent():
        ev.emit('test', ev, 1)
    assert _list == [1]

    ev.set_silent(True)


def test_event_single():
    ev = EventEmitter()

    l = []

    @ev.connect(event='test')
    def on_test_bou(sender):
        l.append(0)

    @ev.connect  # noqa
    def on_test(sender):
        l.append(1)

    ev.emit('test', ev)
    assert l == [0, 1]

    ev.emit('test', ev, single=True)
    assert l == [0, 1, 0]



from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt


def plot_covariance_matrix(covariance_matrix, out_path):
    """
    Function to plot covariance matrix
    :param covariance_matrix:
    :param out_path:
    :return:
    """
    fig, ax = plt.subplots()
    ax.set(title="Data covariance matrix", ylabel='Channel', xlabel='Channel')
    plt.imshow(20 * np.log10(np.abs(covariance_matrix)), vmin=0, vmax=60), plt.colorbar()
    if out_path is not None:
        fig.savefig(Path(out_path).joinpath('_iblqc_.covariance_matrix.png'))
        plt.close(fig)
    else:
        return fig, ax


def plot_whitening_matrix(wrot, whitening_range=32, out_path=None, good_channels=None):
    """
    Function to plot whitening matrix
    :param wrot:
    :param whitening_range:
    :param cond: condition number of whitening matrix
    :return:
    """
    nc = wrot.shape[0]
    good_channels = np.ones(nc, dtype=bool) if good_channels is None else good_channels
    qc_diag = np.zeros((whitening_range * 2, nc))
    cond = np.linalg.cond(wrot[good_channels, :][:, good_channels])
    for i, d in enumerate(np.arange(-whitening_range + 1, whitening_range)):
        diag = np.diag(wrot, d)
        off = int(np.abs(d) / 2)
        qc_diag[i, off:(off + diag.size)] = diag
    fig, ax = plt.subplots(figsize=(12, 4))
    ax.set(title=f"Whitening matrix diagonals, conditioning {cond}", ylabel='Channel', xlabel='Channel')
    plt.imshow(qc_diag, vmin=-.1, vmax=.1, cmap='PuOr'), plt.colorbar(location='bottom')
    if out_path is not None:
        fig.savefig(Path(out_path).joinpath('_iblqc_.whitening_matrix.png'))
        plt.close(fig)
    else:
        return fig, ax



from contextlib import contextmanager
from functools import reduce
import json
import logging
from math import ceil
from pathlib import Path, PurePath
import operator
import os.path as op
import re
from time import perf_counter

from tqdm.auto import tqdm
import numpy as np
from numpy.lib.format import (
    _check_version, _write_array_header, header_data_from_array_1_0, dtype_to_descr)
import cupy as cp

from phylib.io.traces import get_ephys_reader

from .event import emit, connect, unconnect  # noqa

logger = logging.getLogger(__name__)


def prod(iterable):
    return reduce(operator.mul, iterable, 1)


class Bunch(dict):
    """A subclass of dictionary with an additional dot syntax."""
    def __init__(self, *args, **kwargs):
        super(Bunch, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        """Return a new Bunch instance which is a copy of the current Bunch instance."""
        return Bunch(super(Bunch, self).copy())


def copy_bunch(old_bunch):
    """
    A function to copy a bunch object with a deep copy of numpy arrays
    :param old_bunch: Bunch object to be copied
    :return: New Bunch object
    """
    assert type(old_bunch).__name__ == 'Bunch'

    new_bunch = Bunch()
    for key in old_bunch.keys():
        if type(old_bunch[key]) == np.ndarray:
            new_bunch[key] = old_bunch[key].copy()
        else:
            new_bunch[key] = old_bunch[key]

    return new_bunch


def p(x):
    print("shape", x.shape, "mean", "%5e" % x.mean())
    print(x[:2, :2])
    print()
    print(x[-2:, -2:])


def _extend(x, i0, i1, val, axis=0):
    """Extend an array along a dimension and fill it with some values."""
    shape = x.shape
    if x.shape[axis] < i1:
        s = list(x.shape)
        s[axis] = i1 - s[axis]
        x = cp.concatenate((x, cp.zeros(tuple(s), dtype=x.dtype, order='F')), axis=axis)
        assert x.shape[axis] == i1
    s = [slice(None, None, None)] * x.ndim
    s[axis] = slice(i0, i1, 1)
    x[tuple(s)] = val
    for i in range(x.ndim):
        if i != axis:
            assert x.shape[i] == shape[i]
    return x


def is_fortran(x):
    if isinstance(x, np.ndarray):
        return x.flags.f_contiguous
    raise ValueError('`is_fortran` is only implemented for numpy arrays')


def _make_fortran(x):
    if isinstance(x, cp.ndarray):
        x = cp.asnumpy(x)
    return np.asfortranarray(x)


# TODO: design - let's move this to an io module
def read_data(dat_path, offset=0, shape=None, dtype=None, axis=0):
    count = shape[0] * shape[1] if shape and -1 not in shape else -1
    buff = np.fromfile(dat_path, dtype=dtype, count=count, offset=offset)
    if shape and -1 not in shape:
        shape = (-1, shape[1]) if axis == 0 else (shape[0], -1)
    if shape:
        buff = buff.reshape(shape, order='F')
    return buff


# TODO: design - let's move this to an io module
def memmap_binary_file(dat_path, n_channels=None, shape=None, dtype=None, offset=None):
    """Memmap a dat file in FORTRAN order, shape (n_channels, n_samples)."""
    assert dtype is not None
    item_size = np.dtype(dtype).itemsize
    offset = offset if offset else 0
    if shape is None:
        assert n_channels is not None
        n_samples = (op.getsize(str(dat_path)) - offset) // (item_size * n_channels)
        shape = (n_channels, n_samples)
    assert shape
    shape = tuple(shape)
    return np.memmap(str(dat_path), dtype=dtype, shape=shape, offset=offset, order='F')


# TODO: design - move this to cuda/cupy module.
def extract_constants_from_cuda(code):
    r = re.compile(r'const int\s+\S+\s+=\s+\S+.+')
    m = r.search(code)
    if m:
        constants = m.group(0).replace('const int', '').replace(';', '').split(',')
        for const in constants:
            a, b = const.strip().split('=')
            yield a.strip(), int(b.strip())


# TODO: design - move this to cuda/cupy module.
def get_cuda(fn):
    path = Path(__file__).parent / 'cuda' / (fn + '.cu')
    assert path.exists
    code = path.read_text()
    code = code.replace('__global__ void', 'extern "C" __global__ void')
    return code, Bunch(extract_constants_from_cuda(code))


# TODO: design - let's move this to an io module
class LargeArrayWriter(object):
    """Save a large array chunk by chunk, in a binary file with FORTRAN order."""
    def __init__(self, path, dtype=None, shape=None):
        self.path = Path(path)
        self.dtype = np.dtype(dtype)
        self._shape = shape
        assert shape[-1] == -1  # the last axis must be the extendable axis, in FORTRAN order
        assert -1 not in shape[:-1]  # shape may not contain -1 outside the last dimension
        self.fw = open(self.path, 'wb')
        self.extendable_axis_size = 0
        self.total_size = 0

    def append(self, arr):
        # We convert to the requested data type.
        assert arr.flags.f_contiguous  # only FORTRAN order arrays are currently supported
        assert arr.shape[:-1] == self._shape[:-1]
        arr = arr.astype(self.dtype)
        es = arr.shape[-1]
        if arr.flags.f_contiguous:
            arr = arr.T
        # We download the array from the GPU if required.
        # We ensure the array is in FORTRAN order now.
        assert arr.flags.c_contiguous
        if isinstance(arr, cp.ndarray):
            arr = cp.asnumpy(arr)
        arr.tofile(self.fw)
        self.total_size += arr.size
        self.extendable_axis_size += es  # the last dimension, but
        assert prod(self.shape) == self.total_size

    @property
    def shape(self):
        return self._shape[:-1] + (self.extendable_axis_size,)

    def close(self):
        self.fw.close()
        # Save JSON metadata file.
        with open(self.path.with_suffix('.json'), 'w') as f:
            json.dump({'shape': self.shape, 'dtype': str(self.dtype), 'order': 'F'}, f)


# TODO: design - let's move this to an io module
def memmap_large_array(path):
    """Memmap a large array saved by LargeArrayWriter."""
    path = Path(path)
    with open(path.with_suffix('.json'), 'r') as f:
        metadata = json.load(f)
    assert metadata['order'] == 'F'
    dtype = np.dtype(metadata['dtype'])
    shape = metadata['shape']
    return memmap_binary_file(path, shape=shape, dtype=dtype)


# TODO: design - let's move this to an io module
def _npy_header(shape, dtype, order='C'):
    d = {'shape': shape}
    if order == 'C':
        d['fortran_order'] = False
    elif order == 'F':
        d['fortran_order'] = True
    else:
        # Totally non-contiguous data. We will have to make it C-contiguous
        # before writing. Note that we need to test for C_CONTIGUOUS first
        # because a 1-D array is both C_CONTIGUOUS and F_CONTIGUOUS.
        d['fortran_order'] = False

    d['descr'] = dtype_to_descr(dtype)
    return d


# TODO: design - let's move this to an io module
def save_large_array(fp, array, axis=0, desc=None):
    """Save a large, potentially memmapped array, into a NPY file, chunk by chunk to avoid loading
    it entirely in memory."""
    assert axis == 0  # TODO: support other axes
    version = None
    _check_version(version)
    _write_array_header(fp, header_data_from_array_1_0(array), version)
    N = array.shape[axis]
    if N == 0:
        return

    k = int(ceil(float(N) / 100))  # 100 chunks
    assert k >= 1
    for i in tqdm(range(0, N, k), desc=desc):
        chunk = array[i:i + k, ...]
        fp.write(chunk.tobytes())


# TODO: design - let's move this to an io module
class NpyWriter(object):
    def __init__(self, path, shape, dtype, axis=0):
        assert axis == 0  # only concatenation along the first axis is supported right now
        # Only C order is supported at the moment.
        self.shape = shape
        self.dtype = np.dtype(dtype)
        header = _npy_header(self.shape, self.dtype)
        version = None
        _check_version(version)
        self.fp = open(path, 'wb')
        _write_array_header(self.fp, header, version)

    def append(self, chunk):
        if chunk.ndim == len(self.shape):
            assert chunk.shape[1:] == self.shape[1:]
        else:
            assert chunk.shape == self.shape[1:]
        self.fp.write(cp.asnumpy(chunk).tobytes())

    def close(self):
        self.fp.close()


TAPER_LENGTH = 100


class RawDataLoader(object):
    """Class for reading raw data, automatically handles concatenation of multiple datasets as well
    as zero tapering at the boundaries"""
    def __init__(self, data_path, **kwargs):
        """
        :param data_path: Path/s to datasets, string, Pathlib object or list for multiple datasets
        :param kwargs: Arguments for ephys reader
        """
        assert type(data_path) in [str, list] or isinstance(data_path, PurePath), \
            'data_path must be a string, Pathlib path or a list of strings/Pathlib paths'

        # Single dataset case
        if (type(data_path) == str) or isinstance(data_path, PurePath):
            self.multiple_datasets = False
            data_path = Path(data_path)
            assert data_path.is_file(), f'File {data_path} not found'
            self.raw_data = get_ephys_reader(data_path, **kwargs)
            self.n_samples, self.n_channels = self.raw_data.shape
            self.dtype = self.raw_data.dtype
            assert self.n_samples > self.n_channels
            logger.info(f"Loaded data with {self.n_channels} channels, {self.n_samples} samples")

        # Multiple datasets case
        else:
            self.multiple_datasets = True
            self.raw_data = []
            self.n_samples = [0]
            self.n_channels = None
            self.dtype = None
            self.n_datasets = len(data_path)
            for path in data_path:
                assert (type(path) == str) or isinstance(path, PurePath), \
                    'Data paths must be strings or Pathlib paths'
                path = Path(path)
                assert path.is_file(), f'File {path} not found'
                raw_dataset = get_ephys_reader(path, **kwargs)
                n_samples, n_channels = raw_dataset.shape
                logger.info(f"Loaded data with {n_channels} channels, {n_samples} samples")
                if not self.dtype:
                    self.dtype = raw_dataset.dtype
                else:
                    assert self.dtype == raw_dataset.dtype, \
                        'All datasets must have the same data type'
                if not self.n_channels:
                    self.n_channels = n_channels
                else:
                    assert self.n_channels == n_channels, \
                        'All datasets must have the same number of channels'
                self.n_samples.append(n_samples)
                self.raw_data.append(raw_dataset)
            self.n_samples = np.cumsum(np.array(self.n_samples))


    @property
    def total_length(self):
        if not self.multiple_datasets:
            return self.n_samples
        else:
            return self.n_samples[-1]

    @property
    def shape(self):
        return self.total_length, self.n_channels

    @property
    def name(self):
        if not self.multiple_datasets:
            return Path(self.raw_data.name).name
        else:
            return Path(self.raw_data[0].name).name


    # Slice implementation for ease of use
    def __getitem__(self, indices):
        time_start = indices.start or 0
        time_end = indices.stop or self.total_length
        return self.load(time_start, time_end)


    def load(self, time_start, time_end):
        assert type(time_start) == type(time_end) == int
        assert time_start < time_end
        if time_end > self.total_length:
            time_end = self.total_length
        if not self.multiple_datasets:
            return self.raw_data[time_start:time_end]
        else:
            # Get ids of the dataset at the start and end points of the batch
            start_id = np.where(self.n_samples <= time_start)[0][-1]
            end_id = np.where(self.n_samples < time_end)[0][-1]

            # Requested data batch comes from only one dataset
            if start_id == end_id:
                offset = self.n_samples[start_id]
                return self.raw_data[start_id][time_start - offset: time_end - offset]

            # Requested data batch is split across multiple datasets
            else:
                if end_id != start_id + 1:
                    raise NotImplementedError('Unable to load more than two datasets in a batch')

                sub_batch_0 = self.raw_data[start_id][time_start - self.n_samples[start_id]:]
                sub_batch_1 = self.raw_data[end_id][:time_end - self.n_samples[end_id]]

                # To prevent sudden changes at the break point between the two datasets that may
                # lead to artefacts after filtering sine tapering is used

                sub_length = min(TAPER_LENGTH, sub_batch_0.shape[0])
                taper_window = np.sin(np.linspace(
                    0, sub_length / TAPER_LENGTH * np.pi/2, sub_length
                ))[::-1]
                sub_batch_0[-sub_length:] = \
                    (sub_batch_0[-sub_length:] * taper_window.reshape(-1,1)).astype('int16')

                sub_length = min(TAPER_LENGTH, sub_batch_1.shape[0])
                taper_window = np.sin(np.linspace(
                    0, sub_length / TAPER_LENGTH * np.pi/2, sub_length
                ))
                sub_batch_1[:sub_length] = \
                    (sub_batch_0[:sub_length] * taper_window.reshape(-1,1)).astype('int16')

                return np.concatenate((sub_batch_0, sub_batch_1), axis=0)



class DataLoader(object):
    """Class for loading and writing batches in the binary pre-processed data file"""
    def __init__(self, data_path, batch_length, n_channels, scaling_factor, dtype=np.int16):
        """
        :param data_path: Path to binary file
        :param batch_length: Number of time samples in each batch
        :param n_channels: Number of channels
        :param scaling_factor: Scaling factor data is multiplied by before saving
        :param dtype: Data dtype
        """
        self.dtype = dtype
        self.data = np.memmap(data_path, dtype=self.dtype, mode='r+', order='C')
        self.default_batch_length = batch_length
        self.n_channels = n_channels
        self.batch_size = batch_length * n_channels
        self.scaling_factor = scaling_factor

        assert type(self.default_batch_length) == int
        assert type(self.n_channels) == int
        assert self.data.shape[0] % self.batch_size == 0, 'Some batches were only saved partially'

    @property
    def n_batches(self):
        """Returns the number of batches in the dataset"""
        return int(self.data.shape[0] / self.batch_size)

    def load_batch(self, batch_number, batch_length=None, rescale=True):
        """
        Loads a batch and optionally rescales it and move it to the GPU
        :param batch_number: Specifies which batch to load
        :param batch_length: Optional int, length of batch in time samples
        :param rescale: If true, rescales and moves the batch to the GPU
        :return: Loaded batch
        """
        if not batch_length:
            batch_size = self.batch_size
        else:
            batch_size = batch_length * self.n_channels

        assert type(batch_number) == int
        assert batch_number >= 0
        #assert batch_number < self.n_batches, \
        #    f'Batch {batch_number} is out of range for data with {self.n_batches} batches'

        batch_cpu = self.data[batch_number * batch_size : (batch_number + 1) * batch_size].reshape(
            (-1, self.n_channels), order='C'
        )

        if not rescale:
            return np.asfortranarray(batch_cpu)

        batch_gpu = cp.asfortranarray(
            cp.asarray(batch_cpu, dtype=np.float32) / self.scaling_factor
        )

        return batch_gpu

    def write_batch(self, batch_number, batch_data):
        """Saves batch data"""
        assert type(batch_number) == int
        assert batch_number >= 0
        assert batch_number < self.n_batches, \
            f'Batch {batch_number} is out of range for data with {self.n_batches} batches'
        assert batch_data.dtype == self.dtype
        assert batch_data.shape == (self.default_batch_length, self.n_channels)

        self.data[batch_number*self.batch_size:(batch_number+1)*self.batch_size] = batch_data.flatten(order='C')

    def close(self):
        """ Close memmap file, not doing this causes issues on Windows"""
        # TODO: Do this without using internal numpy functions
        self.data._mmap.close()



# TODO: design - this is a nice pythonic-ish mirror of the MATLAB global context,
#              - it might be nicer if it didn't inherit from Bunch (we can be more strict
#              - about what attributes are guaranteed to be there etc.).

# TODO: needs_test - pretty key class and need to make sure we can use it safely
class Context(Bunch):
    def __init__(self, context_path):
        super(Context, self).__init__()
        self.context_path = context_path
        self.intermediate = Bunch()
        self.context_path.mkdir(exist_ok=True, parents=True)
        self.timer = self.load_timer()

    @property
    def metadata_path(self):
        return self.context_path / 'metadata.json'

    def path(self, name, ext='.npy'):
        """Path to an array in the context directory."""
        return self.context_path / (name + ext)

    def read_metadata(self):
        """Read the metadata dictionary from the metadata.json file in the context dir."""
        if not self.metadata_path.exists():
            return Bunch()
        with open(self.metadata_path, 'r') as f:
            return Bunch(json.load(f))

    def write_metadata(self, metadata):
        """Write metadata dictionary in the metadata.json file."""
        with open(self.metadata_path, 'w') as f:
            json.dump(metadata, f, indent=2)

    def read(self, name):
        """Read an array from memory (intermediate object) or from disk."""
        if name not in self.intermediate:
            path = self.path(name)
            # Load a NumPy file.
            if path.exists():
                logger.debug("Loading %s.npy", name)
                # # Memmap for large files.
                # mmap_mode = 'r' if op.getsize(path) > 1e8 else None
                self.intermediate[name] = np.load(path)
            else:
                # Load a value from the metadata file.
                self.intermediate[name] = self.read_metadata().get(name, None)
        return self.intermediate[name]

    def write(self, **kwargs):
        """Write several arrays."""
        # Load the metadata.
        if self.metadata_path.exists():
            metadata = self.read_metadata()
        else:
            metadata = Bunch()
        # Write all variables.
        for k, v in kwargs.items():
            # Transfer GPU arrays to the CPU before saving them.
            if isinstance(v, cp.ndarray):
                logger.debug("Loading %s from GPU.", k)
                v = cp.asnumpy(v)
            if isinstance(v, np.ndarray):
                p = self.path(k)
                overwrite = ' (overwrite)' if p.exists() else ''
                logger.debug("Saving %s.npy%s", k, overwrite)
                np.save(p, np.asfortranarray(v))
            elif v is not None:
                logger.debug("Save %s in the metadata.json file.", k)
                metadata[k] = v
        # Write the metadata file.
        self.write_metadata(metadata)

    def load(self):
        """Load intermediate results from disk."""
        # Load metadata values that are not already loaded in the intermediate dictionary.
        self.intermediate.update(
            {k: v for k, v in self.read_metadata().items() if k not in self.intermediate})
        # Load NumPy arrays that are not already loaded in the intermediate dictionary.
        names = [f.stem for f in self.context_path.glob('*.npy')]
        self.intermediate.update(
            {name: self.read(name) for name in names if name not in self.intermediate})

    def save(self, **kwargs):
        """Save intermediate results to the ctx.intermediate dictionary, and to disk also.

        This has two effects:
        1. variables are available via ctx.intermediate in the current session
        2. In a future session with ctx.load(), these variables will be readily available in
           ctx.intermediate

        """
        for k, v in kwargs.items():
            if v is not None:
                self.intermediate[k] = v
        kwargs = kwargs or self.intermediate
        self.write(**kwargs)

    @contextmanager
    def time(self, name):
        """Context manager to measure the time of a section of code."""
        logger.info("Starting step %s.", name)
        t0 = perf_counter()
        yield
        t1 = perf_counter()
        self.timer[name] = t1 - t0
        self.show_timer(name)
        self.save_timer()

    def show_timer(self, name=None):
        """Display the results of the timer."""
        if name:
            logger.info("Step `{:s}` took {:.2f}s.".format(name, self.timer[name]))
            return
        for name in self.timer.keys():
            self.show_timer(name)
        logger.info("Total time taken was {:.2f}s".format(self.total_time))

    def save_timer(self):
        """Save the results of the time"""
        self.write(timer=self.timer)

    def load_timer(self):
        """Load timer from a previous run"""
        metadata = self.read_metadata()
        if 'timer' in metadata.keys():
            return metadata['timer']
        return {}

    @property
    def total_time(self):
        """Total time taken across all steps"""
        return sum(self.timer.values())


# TODO: design - let's move this to an io module
def load_probe(probe_path):
    """Load a .mat probe file from Kilosort2, or a PRB file (experimental)."""

    # A bunch with the following attributes:
    _required_keys = ('NchanTOT', 'chanMap', 'xc', 'yc', 'kcoords')
    probe = Bunch()
    probe.NchanTOT = 0
    probe_path = Path(probe_path).resolve()

    if probe_path.suffix == '.prb':
        # Support for PRB files.
        contents = probe_path.read_text()
        metadata = {}
        exec(contents, {}, metadata)
        probe.chanMap = []
        probe.xc = []
        probe.yc = []
        probe.kcoords = []
        for cg in sorted(metadata['channel_groups']):
            d = metadata['channel_groups'][cg]
            ch = d['channels']
            pos = d.get('geometry', {})
            probe.chanMap.append(ch)
            probe.NchanTOT += len(ch)
            probe.xc.append([pos[c][0] for c in ch])
            probe.yc.append([pos[c][1] for c in ch])
            probe.kcoords.append([cg for c in ch])
        probe.chanMap = np.concatenate(probe.chanMap).ravel().astype(np.int32)
        probe.xc = np.concatenate(probe.xc)
        probe.yc = np.concatenate(probe.yc)
        probe.kcoords = np.concatenate(probe.kcoords)

    elif probe_path.suffix == '.mat':
        from scipy.io import loadmat
        mat = loadmat(probe_path)
        probe.xc = mat['xcoords'].ravel().astype(np.float64)
        nc = len(probe.xc)
        probe.yc = mat['ycoords'].ravel().astype(np.float64)
        probe.kcoords = mat.get('kcoords', np.zeros(nc)).ravel().astype(np.float64)
        probe.chanMap = (mat['chanMap'] - 1).ravel().astype(np.int32)  # NOTE: 0-indexing in Python
        probe.NchanTOT = len(probe.chanMap)  # NOTE: should match the # of columns in the raw data

    for n in _required_keys:
        assert n in probe.keys()

    return probe


def create_prb(probe):
    chan_map = np.array(probe.chanMap)
    xc, yc = np.array(probe.xc), np.array(probe.yc)
    try:
        bad_channels = np.array(probe.bad_channels)
    except AttributeError:
        bad_channels = np.array([])
    probe_prb = {}
    unique_channel_groups = np.unique(np.array(probe.kcoords))

    for channel_group in unique_channel_groups:
        probe_prb[channel_group] = {}

        channel_group_pos = np.where(probe.kcoords == channel_group)
        group_channels = chan_map[channel_group_pos]
        group_xc = xc[channel_group_pos]
        group_yc = yc[channel_group_pos]

        probe_prb[channel_group]['channels'] = np.setdiff1d(group_channels, bad_channels).tolist()
        geometry = {}

        for c, channel in enumerate(group_channels):
            geometry[channel] = (group_xc[c], group_yc[c])

        probe_prb[channel_group]['geometry'] = geometry
        probe_prb[channel_group]['graph'] = []

    return probe_prb


def plot_dissimilarity_matrices(ccb, ccbsort, plot_widget):
    ccb = cp.asnumpy(ccb)
    ccbsort = cp.asnumpy(ccbsort)

    plot_widget.add_image(
        array=ccb.T,
        plot_pos=0,
        labels={"left": "batches",
                "bottom": "batches",
                "title": "batch to batch distance"
                },
        cmap_style="dissimilarity",
        levels=(0.5, 0.9),
    )

    plot_widget.add_image(
        array=ccbsort.T,
        plot_pos=1,
        labels={"left": "sorted batches",
                "bottom": "sorted batches",
                "title": "AFTER sorting"
                },
        cmap_style="dissimilarity",
        levels=(0.5, 0.9),
    )
    plot_widget.show()


def plot_diagnostics(temporal_comp, spatial_comp, mu, nsp, plot_widget):
    temporal_comp = cp.asnumpy(temporal_comp)
    spatial_comp = cp.asnumpy(spatial_comp)
    mu = cp.asnumpy(mu)
    nsp = cp.asnumpy(nsp)

    plot_widget.add_image(
        array=temporal_comp[:, :, 0].T,
        plot_pos=0,
        labels={"left": "Time (samples)",
                "bottom": "Unit Number",
                "title": "Temporal Components"},
        cmap_style="diagnostic",
        levels=(-0.4, 0.4),
        normalize=False,
    )

    plot_widget.add_image(
        array=spatial_comp[:, :, 0].T,
        plot_pos=1,
        labels={"left": "Channel Number",
                "bottom": "Unit Number",
                "title": "Spatial Components"},
        cmap_style="diagnostic",
        levels=(-0.2, 0.2),
        normalize=False,
    )

    plot_widget.add_curve(
        x_data=np.arange(len(mu)),
        y_data=mu,
        plot_pos=2,
        labels={"left": "Amplitude (arb. units)",
                "bottom": "Unit Number",
                "title": "Unit Amplitudes"},
        y_lim=(0, 100),
    )

    plot_widget.add_scatter(
        x_data=np.log(1+nsp),
        y_data=mu,
        plot_pos=3,
        labels={"left": "Amplitude (arb. units)",
                "bottom": "Spike Count",
                "title": "Amplitude vs. Spike Count"},
        y_lim=(0, 1e2),
        x_lim=(0, np.log(1e5)),
        semi_log_x=True,
        pxMode=True,
        symbol="o",
        size=2,
        pen="w",
    )
    plot_widget.show()



# -*- coding: utf-8 -*-

"""Simple event system."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from contextlib import contextmanager
import logging
import re
from functools import partial

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Event system
#------------------------------------------------------------------------------

class EventEmitter(object):
    """Singleton class that emits events and accepts registered callbacks.

    Example
    -------

    ```python
    class MyClass(EventEmitter):
        def f(self):
            self.emit('my_event', 1, key=2)

    o = MyClass()

    # The following function will be called when `o.f()` is called.
    @o.connect
    def on_my_event(arg, key=None):
        print(arg, key)

    ```

    """

    def __init__(self):
        self.reset()
        self.is_silent = False

    def set_silent(self, silent):
        """Set whether to silence the events."""
        self.is_silent = silent

    def reset(self):
        """Remove all registered callbacks."""
        self._callbacks = []

    def _get_on_name(self, func):
        """Return `eventname` when the function name is `on_<eventname>()`."""
        r = re.match("^on_(.+)$", func.__name__)
        if r:
            event = r.group(1)
        else:
            raise ValueError("The function name should be "
                             "`on_<eventname>`().")
        return event

    @contextmanager
    def silent(self):
        """Prevent all callbacks to be called if events are raised
        in the context manager.
        """
        self.is_silent = not(self.is_silent)
        yield
        self.is_silent = not(self.is_silent)

    def connect(self, func=None, event=None, **kwargs):
        """Register a callback function to a given event.

        To register a callback function to the `spam` event, where `obj` is
        an instance of a class deriving from `EventEmitter`:

        ```python
        @obj.connect
        def on_spam(arg1, arg2):
            pass
        ```

        This is called when `obj.emit('spam', arg1, arg2)` is called.

        Several callback functions can be registered for a given event.

        The registration order is conserved and may matter in applications.

        """
        if func is None:
            return partial(self.connect, event=event, **kwargs)

        # Get the event name from the function.
        if event is None:
            event = self._get_on_name(func)

        # We register the callback function.
        self._callbacks.append((event, func, kwargs))

        return func

    def unconnect(self, *items):
        """Unconnect specified callback functions."""
        self._callbacks = [
            (event, f, kwargs)
            for (event, f, kwargs) in self._callbacks
            if f not in items]

    def emit(self, event, *args, **kwargs):
        """Call all callback functions registered with an event.

        Any positional and keyword arguments can be passed here, and they will
        be forwarded to the callback functions.

        Return the list of callback return results.

        """
        if self.is_silent:
            return
        logger.log(
            5, "Emit %s(%s, %s)", event,
            ', '.join(map(str, args)), ', '.join('%s=%s' % (k, v) for k, v in kwargs.items()))
        # Call the last callback if this is a single event.
        single = kwargs.pop('single', None)
        res = []
        # Put `last=True` callbacks at the end.
        callbacks = [c for c in self._callbacks if not c[-1].get('last', None)]
        callbacks += [c for c in self._callbacks if c[-1].get('last', None)]
        for e, f, k in callbacks:
            if e == event:
                f_name = getattr(f, '__qualname__', getattr(f, '__name__', str(f)))
                logger.log(5, "Callback %s.", f_name)
                res.append(f(*args, **kwargs))
                if single:
                    return res[-1]
        return res


#------------------------------------------------------------------------------
# Global event system
#------------------------------------------------------------------------------

_EVENT = EventEmitter()

emit = _EVENT.emit
connect = _EVENT.connect
unconnect = _EVENT.unconnect
silent = _EVENT.silent
set_silent = _EVENT.set_silent
reset = _EVENT.reset



from math import ceil, erf, log as log_, sqrt
import logging
import os
from os.path import join
from pathlib import Path
import shutil

from tqdm.auto import tqdm
import numba
import numpy as np
import cupy as cp
import cupyx as cpx
from scipy.signal import lfilter
from scipy.sparse import coo_matrix

from .cptools import svdecon, var, mean, free_gpu_memory, convolve_gpu
from .cluster import getClosestChannels
from .learn import getKernels, getMeWtW, mexSVDsmall2
from .preprocess import _is_vect, _make_vect
from .utils import Bunch, NpyWriter, memmap_large_array, LargeArrayWriter

logger = logging.getLogger(__name__)


def log(x):
    if x == 0:
        return -np.inf
    return log_(x)


def my_conv2(x, sig, varargin=None, **kwargs):
    # TODO: Fix so output matches my_conv2_cpu
    # x is the matrix to be filtered along a choice of axes
    # sig is either a scalar or a sequence of scalars, one for each axis to be filtered
    # varargin can be the dimensions to do filtering, if len(sig) != x.shape
    # if sig is scalar and no axes are provided, the default axis is 2
    if sig <= .25:
        return x
    idims = 1
    if varargin is not None:
        idims = varargin
    idims = _make_vect(idims)
    if _is_vect(idims) and _is_vect(sig):
        sigall = sig
    else:
        sigall = np.tile(sig, len(idims))

    for sig, idim in zip(sigall, idims):
        Nd = x.ndim
        x = cp.transpose(x, [idim] + list(range(0, idim)) + list(range(idim + 1, Nd)))
        dsnew = x.shape
        x = cp.reshape(x, (x.shape[0], -1), order='F')

        tmax = ceil(4 * sig)
        dt = cp.arange(-tmax, tmax + 1)
        gaus = cp.exp(-dt ** 2 / (2 * sig ** 2))
        gaus = gaus / cp.sum(gaus)

        y = convolve_gpu(x, gaus, **kwargs)
        y = y.reshape(dsnew, order='F')
        y = cp.transpose(y, list(range(1, idim + 1)) + [0] + list(range(idim + 1, Nd)))
    return y


def my_conv2_cpu(x, sig, varargin=None, **kwargs):
    # (Alternative conv2 function for testing)
    # x is the matrix to be filtered along a choice of axes
    # sig is either a scalar or a sequence of scalars, one for each axis to be filtered
    # varargin can be the dimensions to do filtering, if len(sig) != x.shape
    # if sig is scalar and no axes are provided, the default axis is 2
    if sig <= .25:
        return x
    idims = 1
    if varargin is not None:
        idims = varargin
    idims = _make_vect(idims)
    if _is_vect(idims) and _is_vect(sig):
        sigall = sig
    else:
        sigall = np.tile(sig, len(idims))

    for sig, idim in zip(sigall, idims):
        Nd = x.ndim
        x = cp.transpose(x, [idim] + list(range(0, idim)) + list(range(idim + 1, Nd)))
        dsnew = x.shape
        x = cp.reshape(x, (x.shape[0], -1), order='F')

        tmax = ceil(4 * sig)
        dt = np.arange(-tmax, tmax + 1)
        gaus = np.exp(-dt ** 2 / (2 * sig ** 2))
        gaus = gaus / np.sum(gaus)

        cNorm = lfilter(gaus, np.array([1.]), np.concatenate((np.ones(dsnew[0]), np.zeros(tmax))))
        cNorm = cNorm[tmax:]

        x_n = cp.asnumpy(x)
        x_n = lfilter(gaus, np.array([1.]), np.concatenate((x_n, np.zeros((tmax, dsnew[1])))),
                      axis=0)
        x_n = x_n[tmax:]
        x_n = np.reshape(x_n, dsnew)

        x_n = x_n / cNorm.reshape(-1, 1)
        x = cp.array(x_n)
        x = cp.transpose(x, list(range(1, idim + 1)) + [0] + list(range(idim + 1, Nd)))

    return x


def ccg_slow(st1, st2, nbins, tbin):
    # this function efficiently computes the crosscorrelogram between two sets
    # of spikes (st1, st2), with tbin length each, timelags =  plus/minus nbins
    # and then estimates how refractory the cross-correlogram is, which can be used
    # during merge decisions.

    st1 = cp.sort(st1)  # makes sure spike trains are sorted in increasing order
    st2 = cp.sort(st2)

    dt = nbins * tbin

    N1 = max(1, len(st1))
    N2 = max(1, len(st2))
    T = cp.concatenate((st1, st2)).max() - cp.concatenate((st1, st2)).min()

    # we traverse both spike trains together, keeping track of the spikes in the first
    # spike train that are within dt of spikes in the second spike train

    ilow = 0  # lower bound index
    ihigh = 0  # higher bound index
    j = 0  # index of the considered spike

    K = cp.zeros(2 * nbins + 1)

    # (DEV_NOTES) the while loop below is far too slow as is

    while j <= N2 - 1:  # traverse all spikes in the second spike train

        while (ihigh <= N1 - 1) and (st1[ihigh] < st2[j] + dt):
            ihigh += 1  # keep increasing higher bound until it's OUTSIDE of dt range

        while (ilow <= N1 - 1) and (st1[ilow] <= st2[j] - dt):
            ilow += 1  # keep increasing lower bound until it's INSIDE of dt range

        if ilow > N1 - 1:
            break  # break if we exhausted the spikes from the first spike train

        if st1[ilow] > st2[j] + dt:
            # if the lower bound is actually outside of dt range, means we overshot (there were no
            # spikes in range)
            # simply move on to next spike from second spike train
            j += 1
            continue

        for k in range(ilow, ihigh):
            # for all spikes within plus/minus dt range
            ibin = cp.rint((st2[j] - st1[k]) / tbin).astype(int)  # convert ISI to integer

            K[ibin + nbins] += 1

        j += 1

    irange1 = cp.concatenate((cp.arange(1, nbins // 2), cp.arange(3 * nbins // 2, 2 * nbins)))
    irange2 = cp.arange(nbins - 50, nbins - 10)
    irange3 = cp.arange(nbins + 11, nbins + 50)

    # normalize the shoulders by what's expected from the mean firing rates
    # a non-refractive poisson process should yield 1

    Q00 = cp.sum(K[irange1]) / (len(irange1) * tbin * N1 * N2 / T)
    # do the same for irange 2
    Q01 = cp.sum(K[irange2]) / (len(irange2) * tbin * N1 * N2 / T)
    # compare to the other shoulder
    Q01 = max(Q01, cp.sum(K[irange3]) / (len(irange3) * tbin * N1 * N2 / T))

    R00 = max(mean(K[irange2]), mean(K[irange3]))  # take the biggest shoulder
    R00 = max(R00, mean(K[irange1]))  # compare this to the asymptotic shoulder

    # test the probability that a central area in the autocorrelogram might be refractory
    # test increasingly larger areas of the central CCG

    a = K[nbins]
    K[nbins] = 0

    Qi = cp.zeros(10)
    Ri = cp.zeros(10)

    for i in range(1, 11):
        irange = cp.arange(nbins - i, nbins + i + 1)  # for this central range of the CCG
        # compute the normalised ratio as above. this should be 1 if there is no refractoriness
        Qi0 = cp.sum(K[irange]) / (2 * i * tbin * N1 * N2 / T)
        Qi[i - 1] = Qi0  # save the normalised probability

        n = cp.sum(K[irange]) / 2
        lam = R00 * i

        # log(p) = log(lam) * n - lam - gammaln(n+1)

        # this is tricky: we approximate the Poisson likelihood with a gaussian of equal mean and
        # variance that allows us to integrate the probability that we would see <N spikes in the
        # center of the cross-correlogram from a distribution with mean R00*i spikes

        p = 1 / 2 * (1 + erf((n - lam) / cp.sqrt(2 * lam)))

        Ri[i - 1] = p  # keep track of p for each bin size i

    K[nbins] = a  # restore the center value of the cross-correlogram

    return K, Qi, Q00, Q01, Ri


# NOTE: we get a 50x time improvement with Numba jit of the existing function,
# but we could probably achieve even more by improving the implementation
@numba.jit(nopython=True, cache=False)
def _ccg_old(st1, st2, nbins, tbin):
    # this function efficiently computes the crosscorrelogram between two sets
    # of spikes (st1, st2), with tbin length each, timelags =  plus/minus nbins
    # and then estimates how refractory the cross-correlogram is, which can be used
    # during merge decisions.

    st1 = np.sort(st1)  # makes sure spike trains are sorted in increasing order
    st2 = np.sort(st2)

    dt = nbins * tbin

    # Avoid divide by zero error.
    T = max(1e-10, np.max(np.concatenate((st1, st2))) - np.min(np.concatenate((st1, st2))))
    N1 = max(1, len(st1))
    N2 = max(1, len(st2))

    # we traverse both spike trains together, keeping track of the spikes in the first
    # spike train that are within dt of spikes in the second spike train

    ilow = 0  # lower bound index
    ihigh = 0  # higher bound index
    j = 0  # index of the considered spike

    K = np.zeros(2 * nbins + 1)

    # (DEV_NOTES) the while loop below is far too slow as is

    while j <= N2 - 1:  # traverse all spikes in the second spike train

        while (ihigh <= N1 - 1) and (st1[ihigh] < st2[j] + dt):
            ihigh += 1  # keep increasing higher bound until it's OUTSIDE of dt range

        while (ilow <= N1 - 1) and (st1[ilow] <= st2[j] - dt):
            ilow += 1  # keep increasing lower bound until it's INSIDE of dt range

        if ilow > N1 - 1:
            break  # break if we exhausted the spikes from the first spike train

        if st1[ilow] > st2[j] + dt:
            # if the lower bound is actually outside of dt range, means we overshot (there were no
            # spikes in range)
            # simply move on to next spike from second spike train
            j += 1
            continue

        for k in range(ilow, ihigh):
            # for all spikes within plus/minus dt range
            ibin = np.rint((st2[j] - st1[k]) / tbin)  # convert ISI to integer
            ibin2 = np.asarray(ibin, dtype=np.int64)

            K[ibin2 + nbins] += 1

        j += 1

    irange1 = np.concatenate((np.arange(1, nbins // 2), np.arange(3 * nbins // 2, 2 * nbins)))
    irange2 = np.arange(nbins - 50, nbins - 10)
    irange3 = np.arange(nbins + 11, nbins + 50)

    # normalize the shoulders by what's expected from the mean firing rates
    # a non-refractive poisson process should yield 1

    Q00 = np.sum(K[irange1]) / (len(irange1) * tbin * N1 * N2 / T)
    # do the same for irange 2
    Q01 = np.sum(K[irange2]) / (len(irange2) * tbin * N1 * N2 / T)
    # compare to the other shoulder
    Q01 = max(Q01, np.sum(K[irange3]) / (len(irange3) * tbin * N1 * N2 / T))

    R00 = max(np.mean(K[irange2]), np.mean(K[irange3]))  # take the biggest shoulder
    R00 = max(R00, np.mean(K[irange1]))  # compare this to the asymptotic shoulder

    # test the probability that a central area in the autocorrelogram might be refractory
    # test increasingly larger areas of the central CCG

    a = K[nbins]
    K[nbins] = 0

    Qi = np.zeros(10)
    Ri = np.zeros(10)

    for i in range(1, 11):
        irange = np.arange(nbins - i, nbins + i + 1)  # for this central range of the CCG
        # compute the normalised ratio as above. this should be 1 if there is no refractoriness
        Qi0 = np.sum(K[irange]) / (2 * i * tbin * N1 * N2 / T)
        Qi[i - 1] = Qi0  # save the normalised probability

        n = np.sum(K[irange]) / 2
        lam = R00 * i
        if lam == 0:
            p = np.nan
        else:
            # NOTE: make sure lam is not zero to avoid divide by zero error
            # lam = max(1e-10, R00 * i)

            # log(p) = log(lam) * n - lam - gammaln(n+1)

            # this is tricky: we approximate the Poisson likelihood with a gaussian of equal mean
            # and variance that allows us to integrate the probability that we would see <N spikes
            # in the center of the cross-correlogram from a distribution with mean R00*i spikes

            p = 1 / 2 * (1 + erf((n - lam) / sqrt(2 * lam)))

        Ri[i - 1] = p  # keep track of p for each bin size i

    K[nbins] = a  # restore the center value of the cross-correlogram

    return K, Qi, Q00, Q01, Ri


def ccg_old(st1, st2, nbins, tbin):
    # TODO: move_to_config - There are a lot of parameters in the ccg computation that may need to
    #                      - vary for different setups.
    st1 = cp.asnumpy(st1)
    st2 = cp.asnumpy(st2)
    try:
        return _ccg_old(st1, st2, nbins, tbin)
    except ValueError:
        return 0


def ccg_metrics(st1, st2, nbins, tbin):
    """
    For two arrays of spike times, use the cross-correlogram to estimate the contamination rate
    and the statistical significance that there are fewer spikes in the refractory period
    :param st1: Array of spike times (seconds), numpy or cupy array
    :param st2: Array of spike times (seconds), numpy or cupy array
    :param nbins: Number of time bins either side, int
    :param tbin: Length of each time bin, float
    :return: contam_ratio: Proportion of refractory period violations, float
             p_value: Statistical significance of fewer spikes in the refractory period, float
    """

    K = ccg(st1, st2, nbins, tbin)

    # Indices for the tails of the ccg
    irange1 = np.concatenate((np.arange(1, nbins // 2), np.arange(3 * nbins // 2, 2 * nbins)))

    # Indices for left shoulder of the ccg
    irange2 = np.arange(nbins - 50, nbins - 10)

    # Indices for right shoulder of the ccg
    irange3 = np.arange(nbins + 11, nbins + 50)

    # Estimate the average non-refractory ccg rate by the maximum rate across these ranges
    ccg_rate = max(
        np.mean(K[irange1]),
        np.mean(K[irange2]),
        np.mean(K[irange3]),
    )

    # Set centre of CCG to 0 to avoid double-counted spikes
    K[nbins] = 0

    # test the probability that a central area in the autocorrelogram might be refractory
    # test increasingly larger areas of the central CCG

    contam_rates = np.zeros(10)
    p_values = np.zeros(10)

    for i in range(1, 11):

        irange = np.arange(nbins - i, nbins + i + 1)

        # for this central range of the CCG compute the mean CCG rate
        # as central value is set to 0, divide by 2*i
        contam_rates[i - 1] = np.sum(K[irange]) /  2*i

        n = np.sum(K[irange]) / 2
        lam = ccg_rate * i
        if lam == 0:
            p = 1
        else:
            # NOTE: make sure lam is not zero to avoid divide by zero error
            # lam = max(1e-10, R00 * i)

            # log(p) = log(lam) * n - lam - gammaln(n+1)

            # this is tricky: we approximate the Poisson likelihood with a gaussian of equal mean
            # and variance that allows us to integrate the probability that we would see <N spikes
            # in the center of the cross-correlogram from a distribution with mean R00*i spikes

            p = 1 / 2 * (1 + erf((n - lam) / sqrt(2 * lam)))

        p_values[i - 1] = p  # keep track of p for each bin size i

    # Use the central region that has lowest refractory violation rate
    p_value = np.min(p_values)
    if ccg_rate == 0:
        if np.min(contam_rates) == 0:
            contam_ratio = 0 # CCG is empty so contamination rate set to 0
        else:
            contam_ratio = 1 # Contamination rate is infinite so set to full contamination
    else:
        contam_ratio = np.min(contam_rates) / ccg_rate

    return contam_ratio, p_value


def ccg(st1, st2, nbins, tbin):
    """
    Computes the cross-correlogram for two arrays of spike times
    :param st1: Array of spike times (seconds), numpy or cupy array
    :param st2: Array of spike times (seconds), numpy or cupy array
    :param nbins: Number of time bins either side, int
    :param tbin: Length of each time bin, float
    :return: Cross-correlogram, numpy array
    """
    if (len(st1) == 0) or (len(st2) == 0):
        return np.zeros(2*nbins + 1)

    st1 = cp.asnumpy(st1)
    st2 = cp.asnumpy(st2)

    return _ccg(st1, st2, nbins, tbin)


@numba.jit(nopython=True, cache=False)
def _ccg(st1, st2, nbins, tbin):
    """ JIT compiled ccg function for speed """

    st1 = np.sort(st1)  # makes sure spike trains are sorted in increasing order
    st2 = np.sort(st2)

    dt = nbins * tbin

    # Avoid divide by zero error.
    T = max(1e-10, np.max(np.concatenate((st1, st2))) - np.min(np.concatenate((st1, st2))))
    N1 = max(1, len(st1))
    N2 = max(1, len(st2))

    # we traverse both spike trains together, keeping track of the spikes in the first
    # spike train that are within dt of spikes in the second spike train

    ilow = 0  # lower bound index
    ihigh = 0  # higher bound index
    j = 0  # index of the considered spike

    K = np.zeros(2 * nbins + 1)

    while j <= N2 - 1:  # traverse all spikes in the second spike train

        while (ihigh <= N1 - 1) and (st1[ihigh] < st2[j] + dt):
            ihigh += 1  # keep increasing higher bound until it's OUTSIDE of dt range

        while (ilow <= N1 - 1) and (st1[ilow] <= st2[j] - dt):
            ilow += 1  # keep increasing lower bound until it's INSIDE of dt range

        if ilow > N1 - 1:
            break  # break if we exhausted the spikes from the first spike train

        if st1[ilow] > st2[j] + dt:
            # if the lower bound is actually outside of dt range, means we overshot (there were no
            # spikes in range)
            # simply move on to next spike from second spike train
            j += 1
            continue

        for k in range(ilow, ihigh):
            # for all spikes within plus/minus dt range
            ibin = np.rint((st2[j] - st1[k]) / tbin)  # convert ISI to integer
            ibin2 = np.asarray(ibin, dtype=np.int64)

            K[ibin2 + nbins] += 1

        j += 1

    return K


def clusterAverage(clu, spikeQuantity):
    # get the average of some quantity across spikes in each cluster, given the
    # quantity for each spike
    #
    # e.g.
    # > clusterDepths = clusterAverage(clu, spikeDepths)
    #
    # clu and spikeQuantity must be vector, same size
    #
    # using a super-tricky algorithm for this - when you make a sparse
    # array, the values of any duplicate indices are added. So this is the
    # fastest way I know to make the sum of the entries of spikeQuantity for each of
    # the unique entries of clu
    _, cluInds, spikeCounts = np.unique(clu, return_inverse=True, return_counts=True)

    # summation
    q = coo_matrix((spikeQuantity, (cluInds, np.zeros(len(clu), dtype='int32')))) \
        .toarray().flatten()

    # had sums so dividing by spike counts gives the mean depth of each cluster
    clusterQuantity = q / spikeCounts

    return clusterQuantity


def sort_and_inverse(array):
    """
    Get sorted indices and their inverse for a Numpy array
    :param array: Numpy array
    :return: Sorted indices and their inverse
    """
    perm = np.argsort(array)
    perm_inv = np.zeros_like(perm)
    perm_inv[perm] = np.arange(len(perm))

    return perm, perm_inv


def merge_by_order(array1, array2, ordering1, ordering2, axis=0):
    """
    Concatenate two arrays along given axis according to the given ordering arrays
    This is used to concatenate spike features according to their spike times
    :param array1: Array to be concatenated, fortran order numpy array
    :param array2: Array to be concatenated, fortran order numpy array
    :param ordering1: Ordering for first array, numpy array
    :param ordering2: Ordering for second array, numpy array
    :param axis: Axis along which to concatenate
    :return: Concatenated array, fortran order numpy array
    """
    assert array1.dtype == array2.dtype
    dtype = array1.dtype

    perm1 = np.argsort(ordering1)
    perm2 = np.argsort(ordering2)

    ordering1 = ordering1[perm1]
    ordering2 = ordering2[perm2]

    n1 = array1.shape[axis]
    assert n1 == len(ordering1)
    n2 = array2.shape[axis]
    assert n2 == len(ordering2)

    # Get shape of new array
    new_shape = list(array1.shape)
    new_shape[axis] += array2.shape[axis]

    # Initialise new array
    new_array = np.zeros(new_shape, dtype=dtype, order='F')

    # Get indices for inserting first array
    indices1 = np.arange(n1) + np.searchsorted(ordering2, ordering1)

    # Indices for inserting second array
    mask = np.ones(n1 + n2, dtype='bool')
    mask[indices1] = False
    indices2 = np.where(mask)[0]

    slicer_new = [slice(None)] * new_array.ndim
    slicer_new[axis] = indices1
    slicer_old = [slice(None)] * array1.ndim
    slicer_old[axis] = perm1
    new_array[tuple(slicer_new)] = array1[tuple(slicer_old)]

    slicer_new = [slice(None)] * new_array.ndim
    slicer_new[axis] = indices2
    slicer_old = [slice(None)] * array2.ndim
    slicer_old[axis] = perm2
    new_array[tuple(slicer_new)] = array2[tuple(slicer_old)]

    return new_array


def get_spike_features(feature_path, cluster_id):
    """
    Load spike features for the cluster given by cluster_id
    :param feature_path: Path to spike features folder, Pathlib path
    :param cluster_id: ID of cluster to load
    :return: Spike features, fortran order numpy array
    """

    # Load array as memmap
    memmaped_array = memmap_large_array(feature_path / f'spike_features_{cluster_id}')

    # Convert to numpy array and return
    return np.array(memmaped_array)


def delete_spike_features(feature_path, cluster_id):
    """
    Delete spike features for the cluster given by cluster_id
    :param feature_path: Path to spike features folder, Pathlib path
    :param cluster_id: ID of cluster to delete
    """

    os.remove(feature_path / f'spike_features_{cluster_id}')
    os.remove(feature_path / f'spike_features_{cluster_id}.json')


def write_spike_features(feature_path, cluster_id, dtype, shape, array=None):
    """
    Write array to the spike features on disk for the cluster given by cluster_id
    :param feature_path: Path to spike features
    :param cluster_id: Cluster ID to write to
    :param dtype: Array dtype
    :param shape: Shape of array for array writer
    :param array: Array to write, fortran order numpy array
    """
    writer = LargeArrayWriter(
        feature_path / f'spike_features_{cluster_id}',
        dtype = dtype,
        shape = shape,
        )
    if array is not None:
        writer.append(array)
    writer.close()


def merge_spike_features(feature_path, cluster_1, times_1, cluster_2, times_2, array_shape):
    """
    Merge the spike features for two clusters according to their times, save the new features under
     cluster_1 and delete the old features for cluster_2
    :param feature_path: Path to folder of spike features
    :param cluster_1: Cluster ID of first cluster that will have all the spikes after merging
    :param times_1: Spike times of first cluster
    :param cluster_2: Cluster ID of second cluster that will be empty after merging
    :param times_2: Spike times of second cluster
    :param array_shape: Shape parameter to pass to the array writer
    :return:
    """
    feature_path = Path(feature_path)

    # Load spike features for the clusters
    features_1 = get_spike_features(feature_path, cluster_1)
    features_2 = get_spike_features(feature_path, cluster_2)

    # Combine to get new features with spikes ordered according to their times
    new_features = merge_by_order(features_1,features_2,times_1, times_2, axis=2)

    # Delete the old spike feature files
    delete_spike_features(feature_path, cluster_1)
    delete_spike_features(feature_path, cluster_2)

    # Save new features under cluster_1
    write_spike_features(feature_path, cluster_1, new_features.dtype, array_shape, new_features)

    # Save empty array under cluster_2
    write_spike_features(feature_path, cluster_2, new_features.dtype, array_shape)


def split_features(feature_path, original_cluster, new_cluster, indices, array_shape):
    """
    Split the spike features on disk, spikes in indices are re-assigned from old_cluster to
    new_cluster
    :param feature_path: Path to spike features, Pathlib path
    :param original_cluster: Cluster ID of original cluster, int
    :param new_cluster: Cluster ID of new cluster, int
    :param indices: Spikes to re-assign to new cluster, boolean numpy array
    :param array_shape: Array shape argument to pass to spike feature writer
    """

    feature_path = Path(feature_path)

    # Load spike features of original cluster
    old_features = get_spike_features(feature_path, original_cluster)
    assert old_features.shape[2] == len(indices)

    # New features based on split
    features1 = old_features[:,:,~indices]
    features2 = old_features[:,:,indices]

    delete_spike_features(feature_path, original_cluster)
    assert f'spike_features_{new_cluster}' not in os.listdir(feature_path)

    # Save new features
    write_spike_features(feature_path, original_cluster, old_features.dtype, array_shape,
                         features1)
    write_spike_features(feature_path, new_cluster, old_features.dtype, array_shape,
                         features2)


def find_merges(ctx):
    # this function merges clusters based on template correlation
    # however, a merge is veto-ed if refractory period violations are introduced

    params = ctx.params
    ir = ctx.intermediate

    # TODO: move_to_config
    dt = 1. / 1000  # step size for CCG binning
    nbins = 500  # number of bins used for cross-correlograms
    NchanNear = min(ctx.probe.Nchan, 32)
    Nrank = 3

    # (DEV_NOTES) nbins is not a variable in Marius' code, I include it here to avoid
    # unexplainable, hard-coded constants later

    st3 = ir.st3
    Xsim = ir.simScore  # this is the pairwise similarity score
    Nk = Xsim.shape[0]
    Xsim = Xsim - np.diag(np.diag(Xsim))

    # sort by firing rate first
    nspk = np.zeros(Nk)
    for j in range(Nk):
        # determine total number of spikes in each neuron
        nspk[j] = np.sum(st3[:, 1] == j)

    # we traverse the set of neurons in ascending order of firing rates
    isort = np.argsort(nspk)

    logger.debug('Initialized spike counts.')

    if params.low_memory:
        feature_path = ctx.context_path / 'spike_features'

    for j in tqdm(range(Nk), desc='Finding merges'):
        # find all spikes from this cluster
        s1 = st3[:, 0][st3[:, 1] == isort[j]] / params.fs

        if s1.size != nspk[isort[j]]:
            # this is a check to make sure new clusters are combined correctly into bigger clusters
            # TODO: unclear - don't we want to bail in that case?
            logger.warn('Lost track of spike counts.')

        # sort all the pairs of this neuron, discarding any that have fewer spikes

        uu = Xsim[isort[j], :] * (nspk > s1.size)
        ix = np.argsort(uu)[::-1]
        ccsort = uu[ix]
        ienu = int(np.nonzero(ccsort < .5)[0][0])

        # ccsort = -cp.sort(-Xsim[isort[j]] * (nspk > len(s1)))  # sort in descending order
        # ix = cp.argsort(-Xsim[isort[j]] * (nspk > len(s1)))

        # if ccsort[len(ccsort) - 1] > 0.5:
        #     ienu = len(ccsort)
        # else:
        #     ienu = cp.argmax(ccsort < 0.5)

        # for all pairs above 0.5 correlation

        for k in range(ienu):
            # find the spikes of the pair
            s2 = st3[:, 0][st3[:, 1] == ix[k]] / params.fs

            # Compute contamination ratio and probability the center is refractory from the
            # cross-correlogram. The p_value kicks in when there are very few spikes
            contam_ratio, p_value = ccg_metrics(s1, s2, nbins, dt)

            # TODO: move_to_config
            if (contam_ratio < 0.2) and (p_value < 0.05):  # if both refractory criteria are met
                i = ix[k]
                # now merge j into i and move on
                # simply overwrite all the spikes of neuron j with i (i>j by construction)
                st3[:, 1][st3[:, 1] == isort[j]] = i
                nspk[i] = nspk[i] + nspk[isort[j]]  # update number of spikes for cluster i
                # Update spike features
                if params.low_memory:
                    merge_spike_features(feature_path, int(i), s2, int(isort[j]), s1,
                                         array_shape=(NchanNear, Nrank, -1))
                logger.debug(f'Merged {isort[j]} into {i}')
                # TODO: unclear - the comment below looks important :)
                # YOU REALLY SHOULD MAKE SURE THE PC CHANNELS MATCH HERE
                # break % if a pair is found, we don't need to keep going
                # (we'll revisit this cluster when we get to the merged cluster)
                break

    ctx.save(st3=st3)

    if params.save_temp_files:
        np.save(ctx.context_path / 'temp_splits' / 'st3_merge.npy', st3)


def splitAllClusters(ctx, flag):
    # I call this algorithm "bimodal pursuit"
    # split clusters if they have bimodal projections
    # the strategy is to maximize a bimodality score and find a single vector projection
    # that maximizes it. If the distribution along that maximal projection crosses a
    # bimodality threshold, then the cluster is split along that direction
    # it only uses the PC features for each spike, stored in ir.cProjPC

    params = ctx.params
    probe = ctx.probe
    ir = ctx.intermediate
    Nchan = ctx.probe.Nchan

    wPCA = cp.asarray(ir.wPCA)  # use PCA projections to reconstruct templates when we do splits
    assert wPCA.shape[1] == 3

    # Take intermediate arrays from context.
    st3 = cp.asnumpy(ir.st3)
    cProjPC = ir.cProjPC
    dWU = ir.dWU

    # For the following arrays that will be overwritten by this function, try to get
    # it from a previous call to this function (as it is called twice), otherwise
    # get it from before (without the _s suffix).
    W = ir.get('W_s', ir.W)
    simScore = ir.get('simScore_s', ir.simScore)
    iNeigh = ir.get('iNeigh_s', ir.iNeigh)
    iNeighPC = ir.get('iNeighPC_s', ir.iNeighPC)

    # this is the threshold for splits, and is one of the main parameters users can change
    ccsplit = params.AUCsplit

    Nrank = 3
    NchanNear = min(Nchan, 32)
    Nnearest = min(Nchan, 32)
    sigmaMask = params.sigmaMask

    ik = -1
    Nfilt = W.shape[1]
    nsplits = 0

    # determine what channels each template lives on
    iC, mask, C2C = getClosestChannels(probe, sigmaMask, NchanNear)

    # the waveforms must be aligned to this sample
    nt0min = params.nt0min
    # find the peak abs channel for each template
    iW = np.argmax(np.abs((dWU[nt0min - 1, :, :])), axis=0)

    # keep track of original cluster for each cluster. starts with all clusters being their
    # own origin.
    isplit = np.arange(Nfilt)
    dt = 1. / 1000
    nccg = 0

    if params.low_memory:
        feature_path = ctx.context_path / 'spike_features'

    while ik < Nfilt:
        if ik % 100 == 0:
            # periodically write updates
            logger.info(f'Found {nsplits} splits, checked {ik}/{Nfilt} clusters, nccg {nccg}')
        ik += 1

        isp = (st3[:, 1] == ik)  # get all spikes from this cluster
        nSpikes = isp.sum()
        logger.debug(f"Splitting template {ik}/{Nfilt} with {nSpikes} spikes.")
        free_gpu_memory()

        if nSpikes < 300:
            # TODO: move_to_config
            # do not split if fewer than 300 spikes (we cannot estimate
            # cross-correlograms accurately)
            continue

        ss = st3[isp, 0] / params.fs  # convert to seconds

        # get the PC projections for these spikes
        if params.low_memory:
            clp0 = memmap_large_array(feature_path / f'spike_features_{ik}').T
        else:
            clp0 = cProjPC[isp, :, :]

        clp0 = cp.asarray(clp0, dtype=cp.float32)  # upload to the GPU
        clp0 = clp0.reshape((clp0.shape[0], -1), order='F')
        clp = clp0 - mean(clp0, axis=0) # mean center them

        isp = np.nonzero(isp)[0]

        # subtract a running average, because the projections are NOT drift corrected
        clp = clp - my_conv2(clp, 250, 0)

        # now use two different ways to initialize the bimodal direction
        # the main script calls this function twice, and does both initializations

        if flag:
            u, s, v = svdecon(clp.T)
            #    u, v = -u, -v  # change sign for consistency with MATLAB
            w = u[:, 0]  # initialize with the top PC
        else:
            w = mean(clp0, axis=0)  # initialize with the mean of NOT drift-corrected trace
            w = w / cp.sum(w ** 2) ** 0.5  # unit-normalize

        # initial projections of waveform PCs onto 1D vector
        x = cp.dot(clp, w)
        s1 = var(x[x > mean(x)])  # initialize estimates of variance for the first
        s2 = var(x[x < mean(x)])  # and second gaussian in the mixture of 1D gaussians

        mu1 = mean(x[x > mean(x)])  # initialize the means as well
        mu2 = mean(x[x < mean(x)])
        # and the probability that a spike is assigned to the first Gaussian
        p = mean(x > mean(x))

        # initialize matrix of log probabilities that each spike is assigned to the first
        # or second cluster
        logp = cp.zeros((nSpikes, 2), order='F')

        # do 50 pursuit iteration

        logP = cp.zeros(50)  # used to monitor the cost function

        # TODO: move_to_config - maybe...
        for k in range(50):
            # for each spike, estimate its probability to come from either Gaussian cluster
            logp[:, 0] = -1. / 2 * log(s1) - ((x - mu1) ** 2) / (2 * s1) + log(p)
            logp[:, 1] = -1. / 2 * log(s2) - ((x - mu2) ** 2) / (2 * s2) + log(1 - p)

            lMax = logp.max(axis=1)
            logp = logp - lMax[:, cp.newaxis]  # subtract the max for floating point accuracy
            rs = cp.exp(logp)  # exponentiate the probabilities

            pval = cp.log(cp.sum(rs, axis=1)) + lMax  # get the normalizer and add back the max
            logP[k] = mean(pval)  # this is the cost function: we can monitor its increase

            rs = rs / cp.sum(rs, axis=1)[:, cp.newaxis]  # normalize so that probabilities sum to 1

            p = mean(rs[:, 0])  # mean probability to be assigned to Gaussian 1
            # new estimate of mean of cluster 1 (weighted by "responsibilities")
            mu1 = cp.dot(rs[:, 0], x) / cp.sum(rs[:, 0])
            # new estimate of mean of cluster 2 (weighted by "responsibilities")
            mu2 = cp.dot(rs[:, 1], x) / cp.sum(rs[:, 1])

            s1 = cp.dot(rs[:, 0], (x - mu1) ** 2) / cp.sum(rs[:, 0])  # new estimates of variances
            s2 = cp.dot(rs[:, 1], (x - mu2) ** 2) / cp.sum(rs[:, 1])

            if (k >= 10) and (k % 2 == 0):
                # starting at iteration 10, we start re-estimating the pursuit direction
                # that is, given the Gaussian cluster assignments, and the mean and variances,
                # we re-estimate w
                # these equations follow from the model
                StS = cp.matmul(
                    clp.T, clp * (rs[:, 0] / s1 + rs[:, 1] / s2)[:, cp.newaxis]) / nSpikes
                StMu = cp.dot(clp.T, rs[:, 0] * mu1 / s1 + rs[:, 1] * mu2 / s2) / nSpikes

                # this is the new estimate of the best pursuit direction
                w = cp.linalg.solve(StS.T, StMu)
                w = w / cp.sum(w ** 2) ** 0.5  # which we unit normalize
                x = cp.dot(clp, w)

        # these spikes are assigned to cluster 1
        ilow = rs[:, 0] > rs[:, 1]
        # the mean probability of spikes assigned to cluster 1
        plow = mean(rs[:, 0][ilow])
        phigh = mean(rs[:, 1][~ilow])  # same for cluster 2
        # the smallest cluster has this proportion of all spikes
        nremove = min(mean(ilow), mean(~ilow))

        # did this split fix the autocorrelograms?
        # compute the cross-correlogram between spikes in the putative new clusters
        ilow_cpu = cp.asnumpy(ilow)
        contam_ratio, p_value = ccg_metrics(ss[ilow_cpu], ss[~ilow_cpu], 500, dt)

        # if the CCG has a dip, don't do the split.
        # These thresholds are consistent with the ones from merges.
        # TODO: move_to_config (or at least a single constant so the are the same as the merges)
        if (contam_ratio < 0.25) and (p_value < 0.05):  # if both metrics are below threshold.
            nccg += 1  # keep track of how many splits were voided by the CCG criterion
            continue

        # now decide if the split would result in waveforms that are too similar
        # the reconstructed mean waveforms for putative cluster 1
        # c1 = cp.matmul(wPCA, cp.reshape((mean(clp0[ilow, :], 0), 3, -1), order='F'))
        c1 = cp.matmul(wPCA, mean(clp0[ilow, :], 0).reshape((3, -1), order='F'))
        # the reconstructed mean waveforms for putative cluster 2
        # c2 = cp.matmul(wPCA, cp.reshape((mean(clp0[~ilow, :], 0), 3, -1), order='F'))
        c2 = cp.matmul(wPCA, mean(clp0[~ilow, :], 0).reshape((3, -1), order='F'))

        cc = cp.corrcoef(c1.ravel(), c2.ravel())  # correlation of mean waveforms
        n1 = sqrt(cp.sum(c1 ** 2))  # the amplitude estimate 1
        n2 = sqrt(cp.sum(c2 ** 2))  # the amplitude estimate 2

        r0 = 2 * abs((n1 - n2) / (n1 + n2))

        # if the templates are correlated, and their amplitudes are similar, stop the split!!!

        # TODO: move_to_config
        if (cc[0, 1] > 0.9) and (r0 < 0.2):
            continue

        # finaly criteria to continue with the split: if the split piece is more than 5% of all
        # spikes, if the split piece is more than 300 spikes, and if the confidences for
        # assigning spikes to # both clusters exceeds a preset criterion ccsplit
        # TODO: move_to_config
        if (nremove > 0.05) and (min(plow, phigh) > ccsplit) and (
                min(cp.sum(ilow), cp.sum(~ilow)) > 300):
            # one cluster stays, one goes
            Nfilt += 1

            # the templates for the splits have been estimated from PC coefficients

            # (DEV_NOTES) code below involves multiple CuPy arrays changing shape to accomodate
            # the extra cluster, this could potentially be done more efficiently?

            dWU = cp.concatenate((
                cp.asarray(dWU), cp.zeros((*dWU.shape[:-1], 1), order='F')), axis=2)
            dWU[:, iC[:, iW[ik]], Nfilt - 1] = c2
            dWU[:, iC[:, iW[ik]], ik] = c1

            # the temporal components are therefore just the PC waveforms
            W = cp.asarray(W)
            W = cp.concatenate((W, cp.transpose(cp.atleast_3d(wPCA), (0, 2, 1))), axis=1)
            assert W.shape[1] == Nfilt

            # copy the best channel from the original template
            iW = cp.asarray(iW)
            iW = cp.pad(iW, (0, (Nfilt - len(iW))), mode='constant')
            iW[Nfilt - 1] = iW[ik]
            assert iW.shape[0] == Nfilt

            # copy the provenance index to keep track of splits
            isplit = cp.asarray(isplit)
            isplit = cp.pad(isplit, (0, (Nfilt - len(isplit))), mode='constant')
            isplit[Nfilt - 1] = isplit[ik]
            assert isplit.shape[0] == Nfilt

            st3[isp[ilow_cpu], 1] = Nfilt - 1  # overwrite spike indices with the new index

            # copy similarity scores from the original
            simScore = cp.asarray(simScore)
            simScore = cp.pad(
                simScore, (0, (Nfilt - simScore.shape[0])), mode='constant')
            simScore[:, Nfilt - 1] = simScore[:, ik]
            simScore[Nfilt - 1, :] = simScore[ik, :]
            # copy similarity scores from the original
            simScore[ik, Nfilt - 1] = 1  # set the similarity with original to 1
            simScore[Nfilt - 1, ik] = 1  # set the similarity with original to 1
            assert simScore.shape == (Nfilt, Nfilt)

            # copy neighbor template list from the original
            iNeigh = cp.asarray(iNeigh)
            iNeigh = cp.pad(
                iNeigh, ((0, 0), (0, (Nfilt - iNeigh.shape[1]))), mode='constant')
            iNeigh[:, Nfilt - 1] = iNeigh[:, ik]
            assert iNeigh.shape[1] == Nfilt

            # copy neighbor channel list from the original
            iNeighPC = cp.asarray(iNeighPC)
            iNeighPC = cp.pad(
                iNeighPC, ((0, 0), (0, (Nfilt - iNeighPC.shape[1]))), mode='constant')
            iNeighPC[:, Nfilt - 1] = iNeighPC[:, ik]
            assert iNeighPC.shape[1] == Nfilt

            if params.low_memory:
                # change spike features on disk to reflect the split
                split_features(feature_path, int(ik), int(Nfilt - 1),
                               ilow_cpu, (NchanNear, Nrank, -1))

            # try this cluster again
            # the cluster piece that stays at this index needs to be tested for splits again
            # before proceeding
            ik -= 1
            # the piece that became a new cluster will be tested again when we get to the end
            # of the list
            nsplits += 1  # keep track of how many splits we did
    #         pbar.update(ik)
    # pbar.close()

    logger.info(
        f'Finished splitting. Found {nsplits} splits, checked '
        f'{ik}/{Nfilt} clusters, nccg {nccg}')

    Nfilt = W.shape[1]  # new number of templates
    Nrank = 3
    Nchan = probe.Nchan
    Params = cp.array(
        [0, Nfilt, 0, 0, W.shape[0], Nnearest, Nrank, 0, 0, Nchan, NchanNear, nt0min, 0],
        dtype=cp.float64)  # make a new Params to pass on parameters to CUDA

    # we need to re-estimate the spatial profiles

    # we get the time upsampling kernels again
    Ka, Kb = getKernels(params)
    # we run SVD
    W, U, mu = mexSVDsmall2(Params, dWU, W, iC, iW, Ka, Kb)

    # we re-compute similarity scores between templates
    WtW, iList = getMeWtW(W.astype(cp.float32), U.astype(cp.float32), Nnearest)
    # ir.iList = iList  # over-write the list of nearest templates

    isplit = simScore == 1  # overwrite the similarity scores of clusters with same parent
    simScore = WtW.max(axis=2)
    simScore[isplit] = 1  # 1 means they come from the same parent

    iNeigh = iList[:, :Nfilt]  # get the new neighbor templates
    iNeighPC = iC[:, iW[:Nfilt]]  # get the new neighbor channels

    # for Phy, we need to pad the spikes with zeros so the spikes are aligned to the center of
    # the window
    Wphy = cp.concatenate(
        (cp.zeros((1 + nt0min, Nfilt, Nrank), order='F'), W), axis=0)

    # ir.isplit = isplit  # keep track of origins for each cluster

    if params.save_temp_files:
        np.save(ctx.context_path / 'temp_splits' / 'st3_split.npy', cp.asnumpy(st3))

    #TODO:
    # st3, W, U, mu, simScore, iNeigh, iNeighPC overwrites
    # isplit overwrites or saves
    # Wphy saves
    # iList is repeated and should be deleted (need to check this)

    ctx.save(
        st3=st3,

        W_s=cp.asnumpy(W),
        U_s=cp.asnumpy(U),
        mu_s=cp.asnumpy(mu),
        simScore_s=cp.asnumpy(simScore),
        iNeigh_s=cp.asnumpy(iNeigh),
        iNeighPC_s=cp.asnumpy(iNeighPC),

        Wphy=cp.asnumpy(Wphy),
        iList=cp.asnumpy(iList),
        isplit=cp.asnumpy(isplit),
    )

    if params.save_temp_files:
        ctx.write(st3_s=st3)


def set_cutoff(ctx):
    # after everything else is done, this function takes spike trains and cuts off
    # any noise they might have picked up at low amplitude values
    # We look for bimodality in the amplitude plot, thus setting an individual threshold
    # for each neuron.
    # Also, this function calls "good" and "bad" clusters based on the auto-correlogram

    ir = ctx.intermediate
    params = ctx.params

    st3 = ir.st3  # st3_s1 is saved by the first splitting step
    # cProj = ir.cProj
    # cProjPC = ir.cProjPC

    nbins = 500     # no of bins for CCG
    dt = 1. / 1000  # step size for CCG binning

    Nk = ir.Wphy.shape[1]  # number of templates

    # sort by firing rate first
    good = np.zeros(Nk)
    Ths = np.zeros(Nk)
    est_contam_rate = np.zeros(Nk)

    for j in tqdm(range(Nk), desc='Setting cutoff'):
        ix = np.where(st3[:, 1] == j)[0]  # find all spikes from this neuron
        ss = st3[ix, 0] / params.fs  # convert to seconds
        if ss.size == 0:
            continue  # break if there are no spikes

        vexp = st3[ix, 3]  # vexp is the relative residual variance of the spikes

        Th = params.Th[0]  # start with a high threshold

        fcontamination = 0.1  # acceptable contamination rate
        est_contam_rate[j] = 1

        while Th >= params.Th[1]:
            # continually lower the threshold, while the estimated unit contamination is low
            st = ss[vexp > Th]  # take spikes above the current threshold
            if len(st) == 0:
                Th -= 0.5  # if there are no spikes, we need to keep lowering the threshold
                continue

            # compute the refractory violation metrics using the cross-corellogram
            contam_ratio, p_value = ccg_metrics(st, st, nbins, dt)

            # if the unit is already contaminated, we break, and use the next higher threshold
            if (contam_ratio > fcontamination) or (p_value > 0.05):
                break
            else:
                if (Th == params.Th[0]) and (contam_ratio < 0.05):
                    # only on the first iteration, we consider if the unit starts well isolated
                    # if it does, then we put much stricter criteria for isolation
                    # to make sure we don't settle for a relatively high contamination unit
                    fcontamination = min(0.05, max(0.01, contam_ratio * 2))

                    # if the unit starts out contaminated, we will settle with the higher
                    # contamination rate

                # this unit is good, because we will stop lowering the threshold when it
                # becomes bad
                good[j] = 1
                Th -= 0.5

        # we exited the loop because the contamination was too high. We revert to the higher
        # threshold
        Th += 0.5
        st = ss[vexp > Th]  # take spikes above the current threshold

        # compute the refractory violation metrics using the cross-corellogram
        contam_ratio, p_value = ccg_metrics(st, st, nbins, dt)

        est_contam_rate[j] = contam_ratio  # this score will be displayed in Phy

        Ths[j] = Th  # store the threshold for potential debugging

        # any spikes below the threshold get discarded into a 0-th cluster
        st3[ix[vexp <= Th], 1] = -1

    # we sometimes get NaNs, why? replace with full contamination
    # (DEV_NOTES) this seems to occur when both Qi and max(Q00, Q01) are zero thus when dividing
    # the two to get Q the result is a NaN

    est_contam_rate[np.isnan(est_contam_rate)] = 1

    # remove spikes assigned to the -1 cluster
    ix = st3[:, 1] == -1
    st3 = st3[~ix, :]

    # NOTE: to avoid loading everything into memory, we don't export cProj and cProjPC
    # right now, we'll remove the spikes at the rezToPhy stage.

    # if len(cProj) > 0:
    #     ix = cp.asnumpy(ix)
    #     if len(ix) > cProj.shape[0]:
    #         ix = ix[:cProj.shape[0]]
    #     else:
    #         ix = np.pad(ix, (0, cProj.shape[0] - len(ix)), mode='constant')
    #     assert ix.shape[0] == cProj.shape[0] == cProjPC.shape[0]
    #     cProj = cProj[~ix, :]  # remove their template projections too
    #     cProjPC = cProjPC[~ix, :, :]  # and their PC projections
    #     assert st3.shape[0] == cProj.shape[0] == cProjPC.shape[0]

    ctx.save(
        st3=st3,  # the spikes assigned to -1 have been removed here
        spikes_to_remove=ix,
        # cProj_c=cProj,
        # cProjPC_c=cProjPC,

        est_contam_rate=est_contam_rate,
        Ths=Ths,
        good=good,
    )

    if params.save_temp_files:
        np.save(ctx.context_path / 'temp_splits' / 'st3_cutoff.npy', st3)


def checkClusters(ctx):
    # Checks integrity of clusters. Removes clusters with 0 spikes.

    # 1) List all cluster ids for spikes.
    # 2) Find missing ids
    # 3) Remove these indices from every variable that has n_clusters

    ir = ctx.intermediate
    n_templates = ir.Wphy.shape[1]
    ids = cp.asnumpy(np.unique(ir.st3[:, 1]).astype(int))
    # Check if the max cluster id is equal to the number of cluster ids assigned to spikes.
    if n_templates != len(ids):  # see which cluster ids are missing
        good_units_mask = np.isin(np.arange(n_templates), ids)
        # Remove clusters from fields in `ir` based on `good_units_mask`
        # ir.dWU = ir.dWU[:, :, good_units_mask]
        ir.iNeigh_s = ir.iNeigh_s[:, good_units_mask]
        ir.iNeighPC_s = ir.iNeighPC_s[:, good_units_mask]
        ir.mu_s = ir.mu_s[good_units_mask]
        ir.simScore_s = ir.simScore_s[good_units_mask][:, good_units_mask]
        ir.U_s = ir.U_s[:, good_units_mask, :]
        # ir.UA = ir.UA[:, good_units_mask, :, :]
        # ir.U_a = ir.U_a[:, :, good_units_mask]
        # ir.U_b = ir.U_b[:, :, good_units_mask]
        ir.W_s = ir.W_s[:, good_units_mask, :]
        # ir.WA = ir.WA[:, good_units_mask, :, :]
        # ir.W_a = ir.W_a[:, :, good_units_mask]
        # ir.W_b = ir.W_b[:, :, good_units_mask]
        ir.Wphy = ir.Wphy[:, good_units_mask, :]
        ir.iList = ir.iList[:, good_units_mask]
        ir.isplit = ir.isplit[good_units_mask][:, good_units_mask]
        ir.est_contam_rate = ir.est_contam_rate[good_units_mask]
        ir.Ths = ir.Ths[good_units_mask]
        ir.good = ir.good[good_units_mask]

        # Find empty cluster ids, and for spikes with cluster ids above those indices, subtract 1.
        empty_cl = np.nonzero(~good_units_mask)[0]
        for cl in empty_cl[::-1]:
            logger.debug("Removing empty cluster %d.", cl)
            mislabeled_cl = np.where(ir.st3[:, 1] > cl)[0]
            ir.st3[mislabeled_cl, 1] -= 1

    ctx.ir = ir
    return ctx


# TODO: design - let's split this out into a different module and a class / a few functions
def rezToPhy(ctx, dat_path=None, output_dir=None):
    # pull out results from kilosort's rez to either return to workspace or to
    # save in the appropriate format for the phy GUI to run on. If you provide
    # a output_dir it should be a folder

    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True, parents=True)

    ctx = checkClusters(ctx)  # check clusters integrity

    probe = ctx.probe
    ir = ctx.intermediate
    params = ctx.params

    # spike_times will be in samples, not seconds
    W = ir.Wphy.astype(np.float32)
    Wrot = ir.Wrot
    est_contam_rate = ir.est_contam_rate
    good = ir.good
    Ths = ir.Ths

    st3 = ir.st3

    U = ir.U_s.astype(np.float32)
    iNeigh = ir.iNeigh_s
    iNeighPC = ir.iNeighPC_s
    simScore = ir.simScore_s

    if st3.shape[1] > 4:
        st3 = st3[:, :4]

    isort = np.argsort(st3[:, 0])
    st3 = st3[isort, :]
    # cProj = ir.cProj_c[cp.asnumpy(isort), :]
    # cProjPC = ir.cProjPC_c[cp.asnumpy(isort), :, :]

    fs = os.listdir(output_dir)
    for file in fs:
        if file.endswith('.npy'):
            os.remove(output_dir / file)
    if (output_dir / '.phy').is_dir():
        shutil.rmtree(output_dir / '.phy')

    spike_times = st3[:, 0].astype(np.uint64)
    spike_templates = st3[:, 1].astype(np.uint32)

    # If multiple datasets were run, output the original dataset each spike came from as well as
    # the spike time within the dataset
    if ctx.raw_data.multiple_datasets:
        dataset_times = ctx.raw_data.n_samples
        spike_datasets = np.searchsorted(dataset_times[1:], spike_times, side='right')
        spike_times_corrected = spike_times - dataset_times[spike_datasets]

    # templateFeatures = cProj
    template_feature_inds = iNeigh.astype(np.uint32)
    # pcFeatures = cProjPC
    pc_feature_inds = iNeighPC.astype(np.uint32)

    whiteningMatrix = cp.asnumpy(Wrot) / params.scaleproc
    whiteningMatrixInv = np.linalg.pinv(whiteningMatrix)

    amplitudes = st3[:, 2]

    Nchan = probe.Nchan
    if probe.shank is not None:
        xcoords = probe.x
        ycoords = probe.y
        shanks = probe.shank
    else:
        xcoords = probe.xc
        ycoords = probe.yc
        shanks = None
    chanMap = probe.chanMap
    chanMap0ind = chanMap  # - 1

    nt0, Nfilt = W.shape[:2]

    # (DEV_NOTES) 2 lines below can be combined
    # templates = cp.einsum('ikl,jkl->ijk', U, W).astype(cp.float32)
    # templates = cp.zeros((Nchan, nt0, Nfilt), dtype=np.float32, order='F')
    temp_amps_unscaled = np.zeros(Nfilt, dtype=np.float32)
    templates_writer = NpyWriter(join(output_dir, 'templates.npy'), (Nfilt, nt0, Nchan), np.float32)
    for iNN in tqdm(range(Nfilt), desc="Computing templates"):
        t = np.dot(U[:, iNN, :], W[:, iNN, :].T).T
        templates_writer.append(t)
        t_unw = np.dot(t, whiteningMatrixInv)
        assert t_unw.ndim == 2
        temp_chan_amps = t_unw.max(axis=0) - t_unw.min(axis=0)
        temp_amps_unscaled[iNN] = temp_chan_amps.max()

    templates_writer.close()
    # templates = cp.transpose(templates, (2, 1, 0))  # now it's nTemplates x nSamples x nChannels
    # we include all channels so this is trivial
    templatesInds = np.tile(np.arange(Nfilt), (Nchan, 1))

    # here we compute the amplitude of every template...

    # unwhiten all the templates
    # tempsUnW = cp.einsum('ijk,kl->ijl', templates, whiteningMatrixinv)
    # tempsUnW = cp.zeros(templates.shape, dtype=np.float32, order='F')
    # for t in tqdm(range(templates.shape[0]), desc="Unwhitening the templates"):
    #     tempsUnW[t, :, :] = cp.dot(templates[t, :, :], whiteningMatrixInv)

    # The amplitude on each channel is the positive peak minus the negative
    # temp_chan_amps = tempsUnW.max(axis=1) - tempsUnW.min(axis=1)

    # The template amplitude is the amplitude of its largest channel
    # temp_amps_unscaled = temp_chan_amps.max(axis=1)

    # assign all spikes the amplitude of their template multiplied by their
    # scaling amplitudes
    # temp_amps_unscaled = cp.(temp_amps_unscaled, axis=0).astype(np.float32)
    spike_amps = temp_amps_unscaled[spike_templates] * amplitudes

    # take the average of all spike amps to get actual template amps (since
    # tempScalingAmps are equal mean for all templates)
    ta = clusterAverage(spike_templates, spike_amps)
    tids = np.unique(spike_templates).astype(np.int64)
    temp_amps = np.zeros_like(temp_amps_unscaled, order='F')
    temp_amps[tids] = ta  # because ta only has entries for templates that had at least one spike
    temp_amps = params.gain * temp_amps  # for consistency, make first dimension template number

    # PCs
    ix = ir.spikes_to_remove  # length: number of spikes BEFORE -1 cluster removed

    cProj_shape = ir.cProj.shape
    cProj_shape = (st3.shape[0],) + cProj_shape[1:]

    cProjPC_shape = ir.cProjPC.shape
    cProjPC_shape = (st3.shape[0],) + cProjPC_shape[1:]

    #TODO:
    tfw = NpyWriter(output_dir / 'template_features.npy', cProj_shape, np.float32)
    pcw = NpyWriter(output_dir / 'pc_features.npy', cProjPC_shape, np.float32)

    #isort = cp.asnumpy(isort)
    N = len(ix)  # number of spikes including those assigned to -1
    assert ir.cProj.shape[0] == N
    assert ir.cProjPC.shape[0] == N

    spikes_to_keep = np.nonzero(~ix)[0]  # indices of the spikes to keep in the cProj index space

    # if len(ix) > ir.cProj.shape[0]:
    #     ix = ix[:cProj.shape[0]]
    # else:
    #     ix = np.pad(ix, (0, ir.cProj.shape[0] - len(ix)), mode='constant')
    # assert ix.shape[0] == ir.cProj.shape[0] == ir.cProjPC.shape[0]

    k = int(ceil(float(N) / 100))  # 100 chunks
    assert k >= 1
    for i in tqdm(range(0, N, 100000), desc="Saving template and PC features"):
        # NOTE: cProj and cProjPC still have the spikes assigned to -1 that have yet to be removed

        # spike indices in cProj that need to be kept in this chunk
        ind = spikes_to_keep[isort[i:i + 100000]]

        cProj = ir.cProj[ind]
        cProjPC = ir.cProjPC[ind]

        tfw.append(cProj)
        pcw.append(cProjPC)
    tfw.close()
    pcw.close()
    # with open(, 'wb') as fp:
    #     save_large_array(fp, templateFeatures)
    # cProj = ir.cProj_c[cp.asnumpy(isort), :]
    # cProjPC = ir.cProjPC_c[cp.asnumpy(isort), :, :]

    def _save(name, arr, dtype=None):
        cp.save(output_dir / f'{name}.npy', arr.astype(dtype or arr.dtype))

    if output_dir is not None:
        # units um, dimension (ntimes, ndepths)
        if params.perform_drift_registration:
            _save('drift.um', ir.dshift)
            # units um, dimension (1, ndepths)
            _save('drift_depths.um', ir.yblk[np.newaxis, :])
            batch_size = params.NT / params.fs
            # units secs, dimension (ntimes,)
            _save('drift.times',
                  np.arange(ir.dshift.shape[0]) * batch_size + batch_size / 2)
        _save('spike_times', spike_times)

        # Save two copies as spike_clusters gets modified by Phy during manual curation
        _save('spike_templates', spike_templates, np.uint32)
        _save('spike_clusters', spike_templates, np.uint32)

        if ctx.raw_data.multiple_datasets:
            _save('spike_datasets', spike_datasets)
            _save('spike_times_corrected', spike_times_corrected)

        _save('amplitudes', amplitudes)
        # _save('templates', templates)
        _save('templates_ind', templatesInds)

        chanMap0ind = chanMap0ind.astype(np.int32)

        _save('channel_map', chanMap0ind)
        _save('channel_positions', np.c_[xcoords, ycoords], np.float32)
        _save('channel_labels', probe.channels_labels, np.int8)
        if shanks is not None:
            _save('channel_shanks', shanks, np.int8)

        # _save('template_features', templateFeatures)
        # with open(join(output_dir, 'template_features.npy'), 'wb') as fp:
        #     save_large_array(fp, templateFeatures)
        _save('template_feature_ind', template_feature_inds.T)

        # _save('pc_features', pcFeatures)
        # with open(join(output_dir, 'pc_features.npy'), 'wb') as fp:
        #     save_large_array(fp, pcFeatures)
        _save('pc_feature_ind', pc_feature_inds.T)

        _save('spike_pc_components', ir.wPCA)

        _save('whitening_mat', whiteningMatrix)
        _save('whitening_mat_inv', whiteningMatrixInv)

        _save('thresholds', Ths)

        if 'simScore' in ir:
            similarTemplates = simScore
            _save('similar_templates', similarTemplates)

        est_contam_rate[np.isnan(est_contam_rate)] = 1
        with open(join(output_dir, 'cluster_KSLabel.tsv'), 'w') as f:
            f.write('cluster_id\tKSLabel\n')
            for j in range(len(good)):
                if good[j]:
                    f.write('%d\tgood\n' % j)
                else:
                    f.write('%d\tmua\n' % j)

        # Making a copy with the label group as Phy treats this keyword separately
        with open(join(output_dir, 'cluster_group.tsv'), 'w') as f:
            f.write('cluster_id\tgroup\n')
            for j in range(len(good)):
                if good[j]:
                    f.write('%d\tgood\n' % j)
                else:
                    f.write('%d\tmua\n' % j)

        with open(join(output_dir, 'cluster_ContamPct.tsv'), 'w') as f:
            f.write('cluster_id\tContamPct\n')
            for j in range(len(good)):
                f.write('%d\t%.1f\n' % (j, 100 * est_contam_rate[j]))

        with open(join(output_dir, 'cluster_Amplitude.tsv'), 'w') as f:
            f.write('cluster_id\tAmplitude\n')
            for j in range(len(good)):
                f.write('%d\t%.1f\n' % (j, temp_amps[j]))

        # make params file
        if not os.path.exists(join(output_dir, 'params.py')):
            with open(join(output_dir, 'params.py'), 'w') as f:
                f.write('dat_path = "../%s"\n' % dat_path)
                f.write('n_channels_dat = %d\n' % probe.NchanTOT)
                f.write('dtype = "int16"\n')
                f.write('offset = 0\n')
                f.write('hp_filtered = False\n')
                f.write('sample_rate = %i\n' % params.fs)
                f.write('template_scaling = %.1f\n' % params.templateScaling)



# -*- coding: utf-8 -*-
# flake8: noqa

"""Installation script."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
from pathlib import Path
import re

from setuptools import setup


#------------------------------------------------------------------------------
# Setup
#------------------------------------------------------------------------------

def _package_tree(pkgroot):
    path = op.dirname(__file__)
    subdirs = [op.relpath(i[0], path).replace(op.sep, '.')
               for i in os.walk(op.join(path, pkgroot))
               if '__init__.py' in i[2]]
    return subdirs


readme = (Path(__file__).parent / 'README.md').read_text()


# Find version number from `__init__.py` without executing it.
with (Path(__file__).parent / 'pykilosort/__init__.py').open('r') as f:
    version = re.search(r"__version__ = '([^']+)'", f.read()).group(1)


setup(
    name='pykilosort',
    version=version,
    license="BSD",
    description='Python port of KiloSort 2',
    long_description_content_type="text/markdown",
    long_description=readme,
    author='Cyrille Rossant',
    author_email='cyrille.rossant@gmail.com',
    url='https://github.com/rossant/pykilosort',
    packages=_package_tree('pykilosort'),
    package_dir={'pykilosort': 'pykilosort'},
    package_data={
        'pykilosort': [],
    },
    include_package_data=True,
    keywords='kilosort,spike sorting,electrophysiology,neuroscience',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Framework :: IPython",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'kilosort = pykilosort.gui.launch:launcher'
        ],
    },
)



import logging
from math import ceil
import matplotlib.pyplot as plt

import numpy as np
from scipy.signal import butter
import scipy.stats
import cupy as cp
from tqdm.auto import tqdm

import pykilosort.qc
from .cptools import lfilter, median
from neurodsp.voltage import decompress_destripe_cbin, destripe, detect_bad_channels

logger = logging.getLogger(__name__)


def get_filter_params(fs, fshigh=None, fslow=None):
    if fslow and fslow < fs / 2:
        # butterworth filter with only 3 nodes (otherwise it's unstable for float32)
        return butter(3, (2 * fshigh / fs, 2 * fslow / fs), 'bandpass')
    else:
        # butterworth filter with only 3 nodes (otherwise it's unstable for float32)
        return butter(3, fshigh / fs * 2, 'high')


def gpufilter(buff, chanMap=None, fs=None, fslow=None, fshigh=None, car=True):
    # filter this batch of data after common average referencing with the
    # median
    # buff is timepoints by channels
    # chanMap are indices of the channels to be kep
    # params.fs and params.fshigh are sampling and high-pass frequencies respectively
    # if params.fslow is present, it is used as low-pass frequency (discouraged)

    dataRAW = buff  # .T  # NOTE: we no longer use Fortran order upstream
    assert dataRAW.flags.c_contiguous
    assert dataRAW.ndim == 2
    assert dataRAW.shape[0] > dataRAW.shape[1]
    if chanMap is not None and len(chanMap):
        dataRAW = dataRAW[:, chanMap]  # subsample only good channels
    assert dataRAW.ndim == 2

    # subtract the mean from each channel
    dataRAW = dataRAW - cp.mean(dataRAW, axis=0)  # subtract mean of each channel
    assert dataRAW.ndim == 2

    # CAR, common average referencing by median
    if car:
        # subtract median across channels
        dataRAW = dataRAW - median(dataRAW, axis=1)[:, np.newaxis]

    # set up the parameters of the filter
    filter_params = get_filter_params(fs, fshigh=fshigh, fslow=fslow)

    # next four lines should be equivalent to filtfilt (which cannot be
    # used because it requires float64)
    datr = lfilter(*filter_params, dataRAW, axis=0)  # causal forward filter
    datr = lfilter(*filter_params, datr, axis=0, reverse=True)  # backward
    return datr


# TODO: unclear - Do we really need these, can we not just pick a type for the config?
#               - We can move this complexity into a "config parsing" stage.
def _is_vect(x):
    return hasattr(x, '__len__') and len(x) > 1


def _make_vect(x):
    if not hasattr(x, '__len__'):
        x = np.array([x])
    return x


# TODO: design - can we abstract "running function" out so we don't duplicate most of the code in
#              - my_min and my_max.
def my_min(S1, sig, varargin=None):
    # returns a running minimum applied sequentially across a choice of dimensions and bin sizes
    # S1 is the matrix to be filtered
    # sig is either a scalar or a sequence of scalars, one for each axis to be filtered.
    #  it's the plus/minus bin length for the minimum filter
    # varargin can be the dimensions to do filtering, if len(sig) != x.shape
    # if sig is scalar and no axes are provided, the default axis is 2
    idims = 1
    if varargin is not None:
        idims = varargin
    idims = _make_vect(idims)
    if _is_vect(idims) and _is_vect(sig):
        sigall = sig
    else:
        sigall = np.tile(sig, len(idims))

    for sig, idim in zip(sigall, idims):
        Nd = S1.ndim
        S1 = cp.transpose(S1, [idim] + list(range(0, idim)) + list(range(idim + 1, Nd)))
        dsnew = S1.shape
        S1 = cp.reshape(S1, (S1.shape[0], -1), order='F' if S1.flags.f_contiguous else 'C')
        dsnew2 = S1.shape
        S1 = cp.concatenate(
            (cp.full((sig, dsnew2[1]), np.inf), S1, cp.full((sig, dsnew2[1]), np.inf)), axis=0)
        Smax = S1[:dsnew2[0], :]
        for j in range(1, 2 * sig + 1):
            Smax = cp.minimum(Smax, S1[j:j + dsnew2[0], :])
        S1 = cp.reshape(Smax, dsnew, order='F' if S1.flags.f_contiguous else 'C')
        S1 = cp.transpose(S1, list(range(1, idim + 1)) + [0] + list(range(idim + 1, Nd)))
    return S1


def my_sum(S1, sig, varargin=None):
    # returns a running sum applied sequentially across a choice of dimensions and bin sizes
    # S1 is the matrix to be filtered
    # sig is either a scalar or a sequence of scalars, one for each axis to be filtered.
    #  it's the plus/minus bin length for the summing filter
    # varargin can be the dimensions to do filtering, if len(sig) != x.shape
    # if sig is scalar and no axes are provided, the default axis is 2
    idims = 1
    if varargin is not None:
        idims = varargin
    idims = _make_vect(idims)
    if _is_vect(idims) and _is_vect(sig):
        sigall = sig
    else:
        sigall = np.tile(sig, len(idims))

    for sig, idim in zip(sigall, idims):
        Nd = S1.ndim
        S1 = cp.transpose(S1, [idim] + list(range(0, idim)) + list(range(idim + 1, Nd)))
        dsnew = S1.shape
        S1 = cp.reshape(S1, (S1.shape[0], -1), order='F')
        dsnew2 = S1.shape
        S1 = cp.concatenate(
            (cp.full((sig, dsnew2[1]), 0), S1, cp.full((sig, dsnew2[1]), 0)), axis=0)
        Smax = S1[:dsnew2[0], :]
        for j in range(1, 2 * sig + 1):
            Smax = Smax + S1[j:j + dsnew2[0], :]
        S1 = cp.reshape(Smax, dsnew, order='F')
        S1 = cp.transpose(S1, list(range(1, idim + 1)) + [0] + list(range(idim + 1, Nd)))
    return S1


def whiteningFromCovariance(CC, epsilon=1e-6):
    # function Wrot = whiteningFromCovariance(CC)
    # takes as input the matrix CC of channel pairwise correlations
    # outputs a symmetric rotation matrix (also Nchan by Nchan) that rotates
    # the data onto uncorrelated, unit-norm axes

    # covariance eigendecomposition (same as svd for positive-definite matrix)
    E, D, _ = cp.linalg.svd(CC)
    Di = cp.diag(1. / (D + epsilon) ** .5)
    Wrot = cp.dot(cp.dot(E, Di), E.T)  # this is the symmetric whitening matrix (ZCA transform)
    return Wrot


def whiteningLocal(CC, yc, xc, nRange, epsilon=1e-6):
    # function to perform local whitening of channels
    # CC is a matrix of Nchan by Nchan correlations
    # yc and xc are vector of Y and X positions of each channel
    # nRange is the number of nearest channels to consider
    Wrot = cp.zeros((CC.shape[0], CC.shape[0]))
    for j in range(CC.shape[0]):
        ds = (xc - xc[j]) ** 2 + (yc - yc[j]) ** 2
        ilocal = np.argsort(ds)
        # take the closest channels to the primary channel.
        # First channel in this list will always be the primary channel.
        ilocal = ilocal[:nRange]

        wrot0 = cp.asnumpy(whiteningFromCovariance(CC[np.ix_(ilocal, ilocal)], epsilon=epsilon))
        # the first column of wrot0 is the whitening filter for the primary channel
        Wrot[ilocal, j] = wrot0[:, 0]

    return Wrot


def get_data_covariance_matrix(raw_data, params, probe, nSkipCov=None, preprocessing_function=None):
    """
    Computes the data covariance matrix from a raw data file. Computes from a set of samples from the raw
    data
    :param raw_data:
    :param params:
    :param probe:
    :param nSkipCov: for the orginal kilosort2, number of batches to skip between samples
    :param preprocessing_function: if 'destriping' uses Neuropixel destriping function to apply to the data before
    computing the covariance matrix
    :return:
    """
    preprocessing_function = preprocessing_function or params.preprocessing_function
    if preprocessing_function == 'destriping' and params.normalisation != "original":
        # takes 25 samples of 500ms from 10 seconds to t -25s
        good_channels = probe.good_channels
        CCall = np.zeros((25, probe.Nchan, probe.Nchan))
        t0s = np.linspace(10, raw_data.shape[0] / params.fs - 10, 25)
        # on the second pass, apply destriping to the data
        for icc, t0 in enumerate(tqdm(t0s, desc="Computing covariance matrix")):
            s0 = slice(int(t0 * params.fs), int((t0 + 0.4) * params.fs))
            raw = raw_data[s0][:, :probe.Nchan].T.astype(np.float32) * probe['sample2volt']
            datr = destripe(raw, fs=params.fs, h=probe, channel_labels=probe.channels_labels) / probe['sample2volt']
            assert not (np.any(np.isnan(datr)) or np.any(np.isinf(datr))), "destriping unexpectedly produced NaNs"
            CCall[icc, :, :] = np.dot(datr, datr.T) / datr.shape[1]
            # remove the bad channels from the covariance matrix, those get only divided by their rms
            CCall[icc, :, :] = np.dot(datr, datr.T) / datr.shape[1]
            CCall[icc, ~probe.good_channels, :] = 0
            CCall[icc, :, ~probe.good_channels] = 0
            # we stabilize the covariance matrix this is not trivial:
            # first remove all cross terms belonging to the bad channels
            median_std = np.median(np.diag(CCall[icc, :, :])[good_channels])
            # channels flagged as noisy (label=1) and dead (label=2) are interpolated
            replace_diag = np.interp(np.where(~good_channels)[0], np.where(good_channels)[0], np.diag(CCall[icc, :, :])[good_channels])
            CCall[icc, ~good_channels, ~good_channels] = replace_diag
            # the channels outside of the brain (label=3) are willingly excluded and stay with the high values above
            CCall[icc, probe.channels_labels == 3, probe.channels_labels == 3] = median_std * 1e6
        CC = cp.asarray(np.median(CCall, axis=0))
    else:
        nSkipCov = nSkipCov or params.nSkipCov
        Nbatch = get_Nbatch(raw_data, params)
        NT, NTbuff, ntbuff = (params.NT, params.NTbuff, params.ntbuff)
        nbatches_cov = np.arange(0, Nbatch, nSkipCov).size
        CCall = cp.zeros((nbatches_cov, probe.Nchan, probe.Nchan))
        for icc, ibatch in enumerate(tqdm(range(0, Nbatch, nSkipCov), desc="Computing the whitening matrix")):
            i = max(0, NT * ibatch - ntbuff)
            # WARNING: we no longer use Fortran order, so raw_data is nsamples x NchanTOT
            buff = raw_data[i:i + NTbuff]
            assert buff.shape[0] > buff.shape[1]
            assert buff.flags.c_contiguous
            nsampcurr = buff.shape[0]
            if nsampcurr < NTbuff:
                buff = np.concatenate(
                    (buff, np.tile(buff[nsampcurr - 1], (NTbuff - nsampcurr, 1))), axis=0)
            buff_g = cp.asarray(buff, dtype=np.float32)
            # apply filters and median subtraction
            datr = gpufilter(buff_g, fs=params.fs, fshigh=params.fshigh, chanMap=probe.chanMap)
            # remove buffers on either side of the data batch
            datr = datr[ntbuff: NT + ntbuff]
            assert datr.flags.c_contiguous
            CCall[icc, :, :] = cp.dot(datr.T, datr) / datr.shape[0]
        CC = cp.median(CCall, axis=0)
    return CC


def get_whitening_matrix(raw_data=None, probe=None, params=None, qc_path=None):
    """
    based on a subset of the data, compute a channel whitening matrix
    this requires temporal filtering first (gpufilter)
    :param raw_data:
    :param probe:
    :param params:
    :param kwargs: get_data_covariance_matrix kwargs
    """
    Nchan = probe.Nchan
    CC = get_data_covariance_matrix(raw_data, params, probe)
    logger.info(f"Data normalisation using {params.normalisation} method")
    if params.normalisation in ['whitening', 'original']:
        epsilon = np.mean(np.diag(CC)[probe.good_channels]) * 1e-3 if params.normalisation == 'whitening' else 1e-6
        if params.whiteningRange < np.inf:
            #  if there are too many channels, a finite whiteningRange is more robust to noise
            # in the estimation of the covariance
            whiteningRange = min(params.whiteningRange, Nchan)
            # this function performs the same matrix inversions as below, just on subsets of
            # channels around each channel
            Wrot = whiteningLocal(CC, probe.yc, probe.xc, whiteningRange, epsilon=epsilon)
        else:
            Wrot = whiteningFromCovariance(CC, epsilon=epsilon)
    elif params.normalisation == 'zscore':
        # Do individual channel z-scoring instead of whitening
        Wrot = cp.diag(cp.diag(CC) ** (-0.5))
    elif params.normalisation == 'global_zscore':
        Wrot = cp.eye(CC.shape[0]) * np.median(cp.diag(CC) ** (-0.5))  # same value for all channels
    if qc_path is not None:
        pykilosort.qc.plot_whitening_matrix(Wrot.get(), good_channels=probe.good_channels, out_path=qc_path)
        pykilosort.qc.plot_covariance_matrix(CC.get(), out_path=qc_path)

    Wrot = Wrot * params.scaleproc
    condition_number = np.linalg.cond(cp.asnumpy(Wrot)[probe.good_channels, :][:, probe.good_channels])
    logger.info(f"Computed the whitening matrix cond = {condition_number}.")
    if condition_number > 50:
        logger.warning("high conditioning of the whitening matrix can result in noisy and poor results")
    return Wrot


def get_good_channels(raw_data, probe, params, method='kilosort', **kwargs):
    if method == 'raw_correlations':
        return get_good_channels_raw_correlations(raw_data, probe, params, **kwargs)
    else:
        return get_good_channels_kilosort(raw_data, probe, params)


def get_good_channels_raw_correlations(raw_data, params, probe, t0s=None, return_labels=False):
    """
    Detect bad channels using the method described in IBL whitepaper
    :param raw_data:
    :param params:
    :param probe:
    :param t0s:
    :return:
    """
    if t0s is None:
        t0s = np.linspace(10, raw_data.shape[0] / params.fs - 10, 25)
    channel_labels = np.zeros((probe.Nchan, t0s.size))
    for icc, t0 in enumerate(tqdm(t0s, desc="Auto-detection of noisy channels")):
        s0 = slice(int(t0 * params.fs), int((t0 + 0.4) * params.fs))
        raw = raw_data[s0][:, :probe.Nchan].T.astype(np.float32) * probe['sample2volt']
        channel_labels[:, icc], _ = detect_bad_channels(raw, params.fs)
    channel_labels = scipy.stats.mode(channel_labels, axis=1)[0].squeeze()
    logger.info(f"Detected {np.sum(channel_labels == 1)} dead channels")
    logger.info(f"Detected {np.sum(channel_labels == 2)} noise channels")
    logger.info(f"Detected {np.sum(channel_labels == 3)} uppermost channels outside of the brain")
    if return_labels:
        return channel_labels == 0, channel_labels
    else:
        return channel_labels == 0


def get_good_channels_kilosort(raw_data=None, probe=None, params=None):
    """
    of the channels indicated by the user as good (chanMap)
    further subset those that have a mean firing rate above a certain value
    (default is ops.minfr_goodchannels = 0.1Hz)
    needs the same filtering parameters in ops as usual
    also needs to know where to start processing batches (twind)
    and how many channels there are in total (NchanTOT)
    """
    fs = params.fs
    fshigh = params.fshigh
    fslow = params.fslow
    Nbatch = get_Nbatch(raw_data, params)
    NT = params.NT
    spkTh = params.spkTh
    nt0 = params.nt0
    minfr_goodchannels = params.minfr_goodchannels

    chanMap = probe.chanMap
    NchanTOT = len(chanMap)

    ich = []
    k = 0
    ttime = 0

    # skip every 100 batches
    # TODO: move_to_config - every N batches
    for ibatch in tqdm(range(0, Nbatch, int(ceil(Nbatch / 100))), desc="Finding good channels"):
        i = NT * ibatch
        buff = raw_data[i:i + NT]
        # buff = _make_fortran(buff)
        # NOTE: using C order now
        assert buff.shape[0] > buff.shape[1]
        assert buff.flags.c_contiguous
        if buff.size == 0:
            break

        # Put on GPU.
        buff = cp.asarray(buff, dtype=np.float32)
        assert buff.flags.c_contiguous
        datr = gpufilter(buff, chanMap=chanMap, fs=fs, fshigh=fshigh, fslow=fslow)
        assert datr.shape[0] > datr.shape[1]

        # very basic threshold crossings calculation
        s = cp.std(datr, axis=0)
        datr = datr / s  # standardize each channel ( but don't whiten)
        # TODO: move_to_config (30 sample range)
        mdat = my_min(datr, 30, 0)  # get local minima as min value in +/- 30-sample range

        # take local minima that cross the negative threshold
        xi, xj = cp.nonzero((datr < mdat + 1e-3) & (datr < spkTh))

        # filtering may create transients at beginning or end. Remove those.
        xj = xj[(xi >= nt0) & (xi <= NT - nt0)]

        # collect the channel identities for the detected spikes
        ich.append(xj)
        k += xj.size

        # keep track of total time where we took spikes from
        ttime += datr.shape[0] / fs

    ich = cp.concatenate(ich)

    # count how many spikes each channel got
    nc, _ = cp.histogram(ich, cp.arange(NchanTOT + 1))

    # divide by total time to get firing rate
    nc = nc / ttime

    # keep only those channels above the preset mean firing rate
    igood = cp.asnumpy(nc >= minfr_goodchannels)

    if np.sum(igood) == 0:
        raise RuntimeError("No good channels found! Verify your raw data and parameters.")

    logger.info('Found %d threshold crossings in %2.2f seconds of data.' % (k, ttime))
    logger.info('Found %d/%d bad channels.' % (np.sum(~igood), len(igood)))

    return igood


def get_Nbatch(raw_data, params):
    n_samples = max(raw_data.shape)
    # we assume raw_data as been already virtually split with the requested trange
    return ceil(n_samples / params.NT)  # number of data batches


def destriping(ctx):
    """IBL destriping - multiprocessing CPU version for the time being, although leveraging the GPU
    for the many FFTs performed would probably be quite beneficial """
    probe = ctx.probe
    raw_data = ctx.raw_data
    ir = ctx.intermediate
    wrot = cp.asnumpy(ir.Wrot)
    # get the bad channels
    # detect_bad_channels_cbin
    kwargs = dict(output_file=ir.proc_path, wrot=wrot, nc_out=probe.Nchan, h=probe.h,
                  butter_kwargs={'N': 3, 'Wn': ctx.params.fshigh / ctx.params.fs * 2, 'btype': 'highpass'})

    logger.info("Pre-processing: applying destriping option to the raw data")

    # there are inconsistencies between the mtscomp reader and the flat binary file reader
    # the flat bin reader as an attribute _paths that allows looping on each chunk
    if isinstance(raw_data.raw_data, list):
        for i, rd in enumerate(raw_data.raw_data):
            if i == (len(raw_data.raw_data) - 1):
                ns2add = ceil(raw_data.n_samples[-1] / ctx.params.NT) * ctx.params.NT - raw_data.n_samples[-1]
            else:
                ns2add = 0
            decompress_destripe_cbin(rd.name, ns2add=ns2add, append=i > 0, **kwargs)
    elif getattr(raw_data.raw_data, '_paths', None):
        nstot = 0
        for i, bin_file in enumerate(raw_data.raw_data._paths):
            ns, _ = raw_data.raw_data._mmaps[i].shape
            nstot += ns
            if i == (len(raw_data.raw_data._paths) - 1):
                ns2add = ceil(ns / ctx.params.NT) * ctx.params.NT - ns
            else:
                ns2add = 0
            decompress_destripe_cbin(bin_file, append=i > 0, ns2add=ns2add, **kwargs)
    else:
        assert raw_data.raw_data.n_parts == 1
        ns2add = ceil(raw_data.n_samples / ctx.params.NT) * ctx.params.NT - raw_data.n_samples
        decompress_destripe_cbin(raw_data.raw_data.name, ns2add=ns2add, **kwargs)


def preprocess(ctx):
    # function rez = preprocessDataSub(ops)
    # this function takes an ops struct, which contains all the Kilosort2 settings and file paths
    # and creates a new binary file of preprocessed data, logging new variables into rez.
    # The following steps are applied:
    # 1) conversion to float32
    # 2) common median subtraction
    # 3) bandpass filtering
    # 4) channel whitening
    # 5) scaling to int16 values

    params = ctx.params
    probe = ctx.probe
    raw_data = ctx.raw_data
    ir = ctx.intermediate

    fs = params.fs
    fshigh = params.fshigh
    fslow = params.fslow
    Nbatch = ir.Nbatch
    NT = params.NT
    NTbuff = params.NTbuff
    ntb = params.ntbuff
    Nchan = probe.Nchan

    Wrot = cp.asarray(ir.Wrot)

    logger.info("Loading raw data and applying filters.")

    # weights to combine batches at the edge
    w_edge = cp.linspace(0,1,ntb).reshape(-1, 1)
    datr_prev = cp.zeros((ntb, Nchan), dtype=np.int32)

    with open(ir.proc_path, 'wb') as fw:  # open for writing processed data
        for ibatch in tqdm(range(Nbatch), desc="Preprocessing"):
            # we'll create a binary file of batches of NT samples, which overlap consecutively
            # on params.ntbuff samples
            # in addition to that, we'll read another params.ntbuff samples from before and after,
            # to have as buffers for filtering

            # number of samples to start reading at.
            i = max(0, NT * ibatch - ntb)

            buff = raw_data[i:i + NTbuff]
            if buff.size == 0:
                logger.error("Loaded buffer has an empty size!")
                break  # this shouldn't really happen, unless we counted data batches wrong

            nsampcurr = buff.shape[0]  # how many time samples the current batch has
            if nsampcurr < NTbuff:
                buff = np.concatenate(
                    (buff, np.tile(buff[nsampcurr - 1], (NTbuff - nsampcurr, 1))), axis=0)

            if i == 0:
                bpad = np.tile(buff[0], (ntb, 1))
                buff = np.concatenate((bpad, buff[:NTbuff - ntb]), axis=0)

            # apply filters and median subtraction
            buff = cp.asarray(buff, dtype=np.float32)

            datr = gpufilter(buff, chanMap=probe.chanMap, fs=fs, fshigh=fshigh, fslow=fslow)

            assert datr.flags.c_contiguous

            datr[ntb:2*ntb] = w_edge * datr[ntb:2*ntb] + (1 - w_edge) * datr_prev
            datr_prev = datr[NT + ntb: NT + 2*ntb]

            datr = datr[ntb:ntb + NT, :]  # remove timepoints used as buffers
            datr = cp.dot(datr, Wrot)  # whiten the data and scale by 200 for int16 range
            assert datr.flags.c_contiguous
            if datr.shape[0] != NT:
                raise ValueError(f'Batch {ibatch} processed incorrectly')

            # convert to int16, and gather on the CPU side
            # WARNING: transpose because "tofile" always writes in C order, whereas we want
            # to write in F order.
            datcpu = cp.asnumpy(datr.astype(np.int16))

            # write this batch to binary file
            logger.debug(f"{ir.proc_path.stat().st_size} total, {datr.size * 2} bytes written to file {datcpu.shape} array size")
            datcpu.tofile(fw)
        logger.debug(f"{ir.proc_path.stat().st_size} total")



import numpy as np
import cupy as cp

from pykilosort.learn import mexWtW2


def get_time_lags(array1, array2):
    """
    Takes the dot product of two arrays across all time lags
    :param array1: 1D Numpy array
    :param array2: 1D Numpy array
    :return: 1D Numpy array of dot products
    """
    nt = array1.shape[0]
    n = nt * 2 - 1
    output = np.zeros(n)
    for i in range(n):
        if i < nt:
            output[i] = np.dot(array1[:(i+1)], array2[-(i+1):])
        else:
            output[i] = np.dot(array1[(i+1-nt):], array2[:-(i+1-nt)])
    return output


def mexWtW2_cpu(W1, W2, UtU):
    "Slow cpu version of mex_WtW2 for testing"
    n_times, n_templates = W1.shape
    n = n_times * 2 - 1
    template_corrs = np.zeros((n_templates, n_templates, n))
    for i in range(n_templates):
        for j in range(n_templates):
            template_corrs[i,j] = get_time_lags(W1[:,i], W2[:,j]) * UtU[i,j]
    return template_corrs


def test_mexWtW2():
    n_times = 61
    n_templates = 100

    W1 = np.random.randn(n_times, n_templates)
    W2 = np.random.randn(n_times, n_templates)
    UtU = np.random.random((n_templates, n_templates))

    output_cpu = mexWtW2_cpu(W1, W2, UtU)

    W1 = cp.asarray(W1, dtype=np.float32, order='F')
    W2 = cp.asarray(W2, dtype=np.float32, order='F')
    UtU = cp.asarray(UtU, dtype=np.float32, order='F')

    output = mexWtW2(W1, W2, UtU)
    assert np.allclose(output_cpu, cp.asnumpy(output), atol=1e-05)
